/**
 * @file
 * bitstream reader API header.
 */

#ifndef AVCODEC_GET_BITS_H
#define AVCODEC_GET_BITS_H

#include <stdint.h>
#ifndef NULL
#define NULL ((void*)0)
#endif

static __attribute__((always_inline)) inline uint32_t NEG_USR32(uint32_t a, int8_t s){
    if (__builtin_constant_p(s)) {
        a >>= (32-s);
    } else {
        __asm__ ("shrl %1, %0\n\t"
             : "+r" (a)
             : "ic" ((uint8_t)(-s)));
    }
    return a;
}

static __attribute__((always_inline)) inline uint32_t zero_extend(uint32_t a, int8_t s){
    if (__builtin_constant_p(s)) {
        a = (a << (32-s)) >> (32-s);
    } else {
        __asm__ ("shll %1, %0\n\t"
                 "shrl %1, %0\n\t"
            : "+r" (a)
            : "ic" ((uint8_t)(-s)));
    }
    return a;
}

static __attribute__((always_inline)) inline uint32_t av_bswap32(uint32_t x)
{
    __asm__("bswap   %0" : "+r" (x));
    return x;
}

#ifdef __arm__
static __attribute__((always_inline)) inline uint32_t av_bswap32(uint32_t x)
{
#if HAVE_ARMV6_INLINE
    __asm__("rev %0, %0" : "+r"(x));
#else
    uint32_t t;
    __asm__ ("eor %1, %0, %0, ror #16 \n\t"
             "bic %1, %1, #0xFF0000   \n\t"
             "mov %0, %0, ror #8      \n\t"
             "eor %0, %0, %1, lsr #8  \n\t"
             : "+r"(x), "=&r"(t));
#endif /* HAVE_ARMV6_INLINE */
    return x;
}
#endif

#define AV_RL32(x) ((const uint32_t*)(x))[0]
#define AV_RB32(x) av_bswap32(((const uint32_t*)(x))[0])

/* bit input */
/* buffer, buffer_end and size_in_bits must be present and used by every reader */
typedef struct GetBitContext {
    const unsigned char *buffer;
    unsigned int index;
    unsigned int buffer_size;
} GetBitContext;

#define VLC_TYPE const int16_t

typedef struct VLC {
    int bits;
    VLC_TYPE (*table)[2]; ///< code, bits
    unsigned int table_allocated, table_size;
} VLC;

/* Bitstream reader API docs:
name
    arbitrary name which is used as prefix for the internal variables

gb
    getbitcontext

OPEN_READER(name, gb)
    loads gb into local variables

CLOSE_READER(name, gb)
    stores local vars in gb

UPDATE_CACHE(name, gb)
    refills the internal cache from the bitstream
    after this call at least MIN_CACHE_BITS will be available,

GET_CACHE(name, gb)
    will output the contents of the internal cache, next bit is MSB of 32 or 64 bit (FIXME 64bit)

SHOW_UBITS(name, gb, num)
    will return the next num bits

SHOW_SBITS(name, gb, num)
    will return the next num bits and do sign extension

SKIP_BITS(name, gb, num)
    will skip over the next num bits
    note, this is equivalent to SKIP_CACHE; SKIP_COUNTER

SKIP_CACHE(name, gb, num)
    will remove the next num bits from the cache (note SKIP_COUNTER MUST be called before UPDATE_CACHE / CLOSE_READER)

SKIP_COUNTER(name, gb, num)
    will increment the internal bit counter (see SKIP_CACHE & SKIP_BITS)

LAST_SKIP_CACHE(name, gb, num)
    will remove the next num bits from the cache if it is needed for UPDATE_CACHE otherwise it will do nothing

LAST_SKIP_BITS(name, gb, num)
    is equivalent to LAST_SKIP_CACHE; SKIP_COUNTER

for examples see get_bits, show_bits, skip_bits, get_vlc
*/

#   define MIN_CACHE_BITS 25
#   define OPEN_READER(name, gb)\
        unsigned int name##_index= (gb)->index;\
        unsigned int name##_cache= 0;\

#   define CLOSE_READER(name, gb) (gb)->index= name##_index;
#   define UPDATE_CACHE(name, gb) name##_cache= AV_RB32(((gb)->buffer)+(name##_index>>3)) << (name##_index&0x07);
#   define SKIP_CACHE(name, gb, num) name##_cache <<= (num);

// FIXME name?
#   define SKIP_COUNTER(name, gb, num) name##_index += (num);

#   define SKIP_BITS(name, gb, num)\
        {\
            SKIP_CACHE(name, gb, num)\
            SKIP_COUNTER(name, gb, num)\
        }\

#   define LAST_SKIP_BITS(name, gb, num) SKIP_COUNTER(name, gb, num)
#   define SHOW_UBITS(name, gb, num) NEG_USR32(name##_cache, num)
#   define GET_CACHE(name, gb) ((uint32_t)name##_cache)

static __attribute__((always_inline)) inline int get_bits_count(const GetBitContext *s){
    return s->index;
}

/**
 * shows 1-17 bits.
 */
static __attribute__((always_inline)) inline unsigned int show_bits(GetBitContext *s, int n){
    unsigned int re_index=s->index, re_cache= AV_RB32(s->buffer+(re_index>>3)) << (re_index&0x07);
    unsigned int tmp= NEG_USR32(re_cache, n);
    return tmp;
}

static __attribute__((always_inline)) inline unsigned int show_bits_le(GetBitContext *s, int n){
    unsigned int re_index=s->index, re_cache= AV_RL32(s->buffer+(re_index>>3)) >> (re_index&0x07);
    unsigned int tmp= zero_extend(re_cache, n);
    return tmp;
}

static __attribute__((always_inline)) inline void skip_bits(GetBitContext *s, int n){
	s->index += n;
}

/**
 * reads 1-17 bits.
 */
static __attribute__((always_inline)) inline unsigned int get_bits(GetBitContext *s, int n){
    register int tmp = show_bits(s, n);
    s->index += n;
    return tmp;
}

static __attribute__((always_inline)) inline unsigned int get_bits_le(GetBitContext *s, int n){
    register int tmp = show_bits_le(s, n);
    s->index += n;
    return tmp;
}

static __attribute__((always_inline)) inline unsigned int show_bits1(GetBitContext *s){
    unsigned int re_index= s->index;
    uint8_t result= s->buffer[re_index>>3] << (re_index&0x07);
    result>>= 8 - 1;
    return result;
}

static __attribute__((always_inline)) inline unsigned int get_bits1(GetBitContext *s){
    register int tmp = show_bits1(s);
    s->index++;
    return tmp;
}

/**
 * init GetBitContext.
 * @param buffer bitstream buffer, must be FF_INPUT_BUFFER_PADDING_SIZE bytes larger then the actual read bits
 * because some optimized bitstream readers read 32 or 64 bit at once and could read over the end
 * @param bit_size the size of the buffer in bits
 *
 * While GetBitContext stores the buffer size, for performance reasons you are
 * responsible for checking for the buffer end yourself (take advantage of the padding)!
 */
static __attribute__((always_inline)) inline void init_get_bits(GetBitContext *s, const uint8_t *buffer, uint32_t buffer_size)
{
    s->buffer= buffer;
    s->buffer_size= buffer_size;
    s->index=0;
}

static inline void align_get_bits(GetBitContext *s)
{
    int n= (-get_bits_count(s)) & 7;
    if(n) skip_bits(s, n);
}

/**
 *
 * If the vlc code is invalid and max_depth=1, then no bits will be removed.
 * If the vlc code is invalid and max_depth>1, then the number of bits removed
 * is undefined.
 */
#define GET_VLC(code, name, gb, table, bits, max_depth)\
{\
    int n, nb_bits;\
    unsigned int index;\
\
    index= SHOW_UBITS(name, gb, bits);\
    code = table[index][0];\
    n    = table[index][1];\
    if (n < 0){\
        LAST_SKIP_BITS(name, gb, bits)\
        UPDATE_CACHE(name, gb)\
        nb_bits = -n;\
        index= SHOW_UBITS(name, gb, nb_bits) + code;\
        code = table[index][0];\
        n    = table[index][1];\
    }\
    SKIP_BITS(name, gb, n)\
}

static inline int get_bits_left(GetBitContext *gb)
{
    return (gb->buffer_size << 3) - get_bits_count(gb);
}

#endif /* AVCODEC_GET_BITS_H */
