/*
 * MPEG Audio decoder
 * Copyright (c) 2001, 2002 Fabrice Bellard
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file
 * MPEG Audio decoder.
 */

#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include "get_bits.h"
#include "mp3.h"
extern const VLC mp3_huff_vlc[16];

/* Copy of the av_log() API from ffmpeg, minus the context parameter. */
#define AV_LOG_FATAL     8
#define AV_LOG_ERROR    16
#define AV_LOG_WARNING  24
#define AV_LOG_INFO     32
#define AV_LOG_VERBOSE  40
#define AV_LOG_DEBUG    48

#ifdef __GNUC__
void av_log(int level, const char *fmt, ...) __attribute__ ((__format__ (__printf__, 2, 3)));
#else
void av_log(int level, const char *fmt, ...);
#endif

/* Useful macros. */
#define FFMAX(a,b) ((a) > (b) ? (a) : (b))
#define FFMIN(a,b) ((a) > (b) ? (b) : (a))
#define FFABS(a) ((a) >= 0 ? (a) : (-(a)))
#define FF_ARRAY_ELEMS(a) (sizeof(a) / sizeof((a)[0]))
#define DECLARE_ALIGNED(n,t,v) t __attribute__ ((aligned (n))) v

#ifndef FFTCOMPLEX_T_DEFINED
typedef struct FFTComplex {
    float re, im;
} FFTComplex;
#define FFTCOMPLEX_T_DEFINED
#endif

typedef union _f32 {
    float f;
    uint32_t i;
} _f32;

void dct32_c(float *out, float *in);
void dct32_sse(float *out, float *in);
void dct32_neon(float *out, float *in);
void imdct36_c(float *out, float *in);
void imdct36_sse(float *out, float *in);
void imdct36_neon(float *out, float *in);

/*
 * TODO:
 *  - test lsf / mpeg25 extensively.
 */
#define FIXHR(x)       ((float)(x))

/****************/
#define HEADER_SIZE 4
void ff_mpa_synth_init(void);
void ff_mpa_synth_filter(float *synth_buf_ptr, int *synth_buf_offset, int *dither_state,
                         float *samples, float sb_samples[SBLIMIT]);

/* vlc structure for decoding layer 3 huffman tables */
static const uint8_t huff_quad_vlc_table1[128][2] = {
    { 0x0b, 6 }, { 0x0b, 6 }, { 0x0f, 6 }, { 0x0f, 6 }, { 0x0d, 6 }, { 0x0d, 6 }, { 0x0e, 6 }, { 0x0e, 6 }, 
    { 0x07, 6 }, { 0x07, 6 }, { 0x05, 6 }, { 0x05, 6 }, { 0x09, 5 }, { 0x09, 5 }, { 0x09, 5 }, { 0x09, 5 }, 
    { 0x06, 5 }, { 0x06, 5 }, { 0x06, 5 }, { 0x06, 5 }, { 0x03, 5 }, { 0x03, 5 }, { 0x03, 5 }, { 0x03, 5 }, 
    { 0x0a, 5 }, { 0x0a, 5 }, { 0x0a, 5 }, { 0x0a, 5 }, { 0x0c, 5 }, { 0x0c, 5 }, { 0x0c, 5 }, { 0x0c, 5 }, 
    { 0x02, 4 }, { 0x02, 4 }, { 0x02, 4 }, { 0x02, 4 }, { 0x02, 4 }, { 0x02, 4 }, { 0x02, 4 }, { 0x02, 4 }, 
    { 0x01, 4 }, { 0x01, 4 }, { 0x01, 4 }, { 0x01, 4 }, { 0x01, 4 }, { 0x01, 4 }, { 0x01, 4 }, { 0x01, 4 }, 
    { 0x04, 4 }, { 0x04, 4 }, { 0x04, 4 }, { 0x04, 4 }, { 0x04, 4 }, { 0x04, 4 }, { 0x04, 4 }, { 0x04, 4 }, 
    { 0x08, 4 }, { 0x08, 4 }, { 0x08, 4 }, { 0x08, 4 }, { 0x08, 4 }, { 0x08, 4 }, { 0x08, 4 }, { 0x08, 4 }, 
    { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, 
    { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, 
    { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, 
    { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, 
    { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, 
    { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, 
    { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, 
    { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }, { 0x00, 1 }
};

static const uint8_t huff_quad_vlc_table2[16][2] = {
    { 0x0f, 4 }, { 0x0e, 4 }, { 0x0d, 4 }, { 0x0c, 4 }, { 0x0b, 4 }, { 0x0a, 4 }, { 0x09, 4 }, { 0x08, 4 }, 
    { 0x07, 4 }, { 0x06, 4 }, { 0x05, 4 }, { 0x04, 4 }, { 0x03, 4 }, { 0x02, 4 }, { 0x01, 4 }, { 0x00, 4 }
};


/* computed from band_size_long */
static uint16_t band_index_long[9][23];
#define TABLE_4_3_SIZE (8191 + 16)*4
static int8_t   table_4_3_exp[TABLE_4_3_SIZE];
static uint32_t table_4_3_value[TABLE_4_3_SIZE];
static float exp_table_float[512];

static const float mp3_pow2sf_frac[4] = { 
    1.0000000, 1.1892071, 1.4142135, 1.6817928
};

static const float pow4_3_tab[16] = {
     0.0000000,  1.0000000,
     2.5198421,  4.3267487,
     6.3496042,  8.5498797,
    10.9027236, 13.3905183,
    16.0000000, 18.7207544,
    21.5443469, 24.4637810,
    27.4731418, 30.5673509,
    33.7419917, 36.9931811,
};

static inline float fft_ldexpf1(int n)
{
    _f32 tmp;
    /* normal result */
    n += 0x7f;
    tmp.i = (n << 23);
    return tmp.f;
}

/* Constants for cbrtf(x) -- single-precision cube-root of x. */
static const unsigned
B1 = 1419915941; /* B1 = (169+1/3-0.0661629)*2**23 */

float fft_cbrtf(float x)
{
    _f32 tmp;
    float r, t;
    uint32_t sign = (x < 0.0f);

    tmp.f = fabsf(x); /* x <- |x| */

    if(!tmp.i || tmp.i>=0x7f800000) return(x); /* cbrt(0,NaN,INF) is itself */

    /* rough cbrt to 5 bits */
    tmp.i = (B1-tmp.i/3);
    t = tmp.f;

    /* new cbrt to 23 bits */
    r = t*t*t;
    t *= 0.333333333f * (4.0f - x*r);
    r = t*t*t;
    t *= 0.333333333f * (4.0f - x*r);
    r = t*t*t;
    t *= 0.333333333f * (4.0f - x*r);

    /* convert 1/cbrtf(x) to cbrtf(x) */
    t *= t;
    t *= fabsf(x);

    /* retore the sign bit */
    if (sign) t = -t;
    return(t);
}

float frexpf(float x, int *eptr)
{
    _f32 f;
    f.f = fabsf(x);
	*eptr = 0;
	if(f.i>=0x7f800000||(f.i<0x00800000)) return x;	/* 0,inf,nan,denormals */
	*eptr += (f.i>>23)-127;
	f.i = (f.i&0x007fffff)|0x3f800000;
    if (x < 0.0f) f.f = -f.f;
	return f.f;
}

static void mpegaudio_tableinit(void)
{
    int i, exponent;
    for (i = 1; i < TABLE_4_3_SIZE; i++) {
        float value = i / 4;
        float f, fm;
        int e, m;
        f  = value * fft_cbrtf(value) * mp3_pow2sf_frac[(i & 3)];
        fm = frexpf(f, &e);
        m  = lrintf(fm * (1LL << 31));
        e += FRAC_BITS - 31 + 5 - 100;

        /* normalized to FRAC_BITS */
        table_4_3_value[i] =  m;
        table_4_3_exp[i]   = -e;
    }
    for (exponent = 0; exponent < 128; exponent++) {
        //float f = fft_ldexpf1((exponent>>2) - 100 + FRAC_BITS + 5) * mp3_pow2sf_frac[exponent & 3]; 
        float f = fft_ldexpf1(exponent - 100 + FRAC_BITS + 5); 
        exp_table_float[4*exponent+0] = f * mp3_pow2sf_frac[0];
        exp_table_float[4*exponent+1] = f * mp3_pow2sf_frac[1];
        exp_table_float[4*exponent+2] = f * mp3_pow2sf_frac[2];
        exp_table_float[4*exponent+3] = f * mp3_pow2sf_frac[3];
    }
}

/* intensity stereo coef table */
static float is_table[2][16] = {
    { 0.000000, 0.211325, 0.366025, 0.500000, 0.633975, 0.788675, 1.000000, 0.000000,
      0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000 },
    { 1.000000, 0.788675, 0.633975, 0.500000, 0.366025, 0.211325, 0.000000, 0.000000,
      0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000 } 
};
static float is_table_lsf0[2][16] = {
  {  1.0, 0.84089641525, 1.0, 0.70710678118, 1.0, 0.5946035575, 1.0, 0.5, 1.0, 0.42044820762, 1.0, 0.35355339059, 1.0, 0.29730177875, 1.0, 0.25 },
  {  1.0, 1.0, 0.84089641525, 1.0, 0.70710678118, 1.0, 0.5946035575, 1.0, 0.5, 1.0, 0.42044820762, 1.0, 0.35355339059, 1.00, 0.29730177875, 1.0 }
};
static float is_table_lsf1[2][16] = {
  { 1.0, 0.7071068, 1.0, 0.5, 1.0, 0.3535534, 1.0, 0.25, 1.0, 0.1767767, 1.0, 0.125, 1.0, 0.0883883, 1.0, 0.0625 },
  { 1.0, 1.0, 0.7071068, 1.0, 0.5, 1.0, 0.3535534, 1.0, 0.25, 1.0, 0.1767767, 1.0, 0.125, 1.0, 0.0883883, 1.0 }
};
static float csa_table_float[8][4];
static float mdct_win[8][36];

static int16_t division_tab3[1<<6 ];
static int16_t division_tab5[1<<8 ];
static int16_t division_tab9[1<<11];

static int16_t * const division_tabs[4] = {
    division_tab3, division_tab5, NULL, division_tab9
};

/* lower 2 bits: modulo 3, higher bits: shift */
static uint16_t scale_factor_modshift[64];
/* [i][j]:  2^(-j/3) * FRAC_ONE * 2^(i+2) / (2^(i+2) - 1) */
static int32_t scale_factor_mult[15][3];
/* mult table for layer 2 group quantization */
static const uint32_t scale_factor_mult2[3][3] = {
    { 11184811, 8877390, 7045989 }, /* 3 steps */
    {  6710886, 5326434, 4227594 }, /* 5 steps */
    {  3728270, 2959130, 2348663 }, /* 9 steps */
};

/*******************************************************/
/* layer 2 tables */

static const uint8_t ff_mpa_sblimit_table[5] = { 27 , 30 , 8, 12 , 30 };

static const uint16_t ff_mpa_quant_steps[17] = {
    3,     5,    7,    9,    15,
    31,    63,  127,  255,   511,
    1023,  2047, 4095, 8191, 16383,
    32767, 65535
};

/* we use a negative value if grouped */
static const int8_t ff_mpa_quant_bits[17] = {
    -5,  -7,  3, -10, 4,
     5,  6,  7,  8,  9,
    10, 11, 12, 13, 14,
    15, 16
};

/*******************************************************/
/* layer 3 tables */

/* layer3 scale factor size */
static const uint8_t slen_table[2][16] = {
    { 0, 0, 0, 0, 3, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4 },
    { 0, 1, 2, 3, 0, 1, 2, 3, 1, 2, 3, 1, 2, 3, 2, 3 },
};

/* number of lsf scale factors for a given size */
static const uint8_t lsf_nsf_table[6][3][4] = {
    { {  6,  5,  5, 5 }, {  9,  9,  9, 9 }, {  6,  9,  9, 9 } },
    { {  6,  5,  7, 3 }, {  9,  9, 12, 6 }, {  6,  9, 12, 6 } },
    { { 11, 10,  0, 0 }, { 18, 18,  0, 0 }, { 15, 18,  0, 0 } },
    { {  7,  7,  7, 0 }, { 12, 12, 12, 0 }, {  6, 15, 12, 0 } },
    { {  6,  6,  6, 3 }, { 12,  9,  9, 6 }, {  6, 12,  9, 6 } },
    { {  8,  8,  5, 0 }, { 15, 12,  9, 0 }, {  6, 18,  9, 0 } },
};

static const uint8_t mpa_huff_data[32][2] = {
{ 0, 0 },
{ 1, 0 },
{ 2, 0 },
{ 3, 0 },
{ 0, 0 },
{ 4, 0 },
{ 5, 0 },
{ 6, 0 },
{ 7, 0 },
{ 8, 0 },
{ 9, 0 },
{ 10, 0 },
{ 11, 0 },
{ 12, 0 },
{ 0, 0 },
{ 13, 0 },
{ 14, 1 },
{ 14, 2 },
{ 14, 3 },
{ 14, 4 },
{ 14, 6 },
{ 14, 8 },
{ 14, 10 },
{ 14, 13 },
{ 15, 4 },
{ 15, 5 },
{ 15, 6 },
{ 15, 7 },
{ 15, 8 },
{ 15, 9 },
{ 15, 11 },
{ 15, 13 },
};

/* band size tables */
static const uint8_t band_size_long[9][22] = {
{ 4, 4, 4, 4, 4, 4, 6,  6,  8,  8, 10, 12, 16, 20, 24, 28, 34, 42, 50, 54,  76, 158 }, /* 44100 */
{ 4, 4, 4, 4, 4, 4, 6,  6,  6,  8, 10, 12, 16, 18, 22, 28, 34, 40, 46, 54,  54, 192 }, /* 48000 */
{ 4, 4, 4, 4, 4, 4, 6,  6,  8, 10, 12, 16, 20, 24, 30, 38, 46, 56, 68, 84, 102,  26 }, /* 32000 */
{ 6, 6, 6, 6, 6, 6, 8, 10, 12, 14, 16, 20, 24, 28, 32, 38, 46, 52, 60, 68,  58,  54 }, /* 22050 */
{ 6, 6, 6, 6, 6, 6, 8, 10, 12, 14, 16, 18, 22, 26, 32, 38, 46, 52, 64, 70,  76,  36 }, /* 24000 */
{ 6, 6, 6, 6, 6, 6, 8, 10, 12, 14, 16, 20, 24, 28, 32, 38, 46, 52, 60, 68,  58,  54 }, /* 16000 */
{ 6, 6, 6, 6, 6, 6, 8, 10, 12, 14, 16, 20, 24, 28, 32, 38, 46, 52, 60, 68,  58,  54 }, /* 11025 */
{ 6, 6, 6, 6, 6, 6, 8, 10, 12, 14, 16, 20, 24, 28, 32, 38, 46, 52, 60, 68,  58,  54 }, /* 12000 */
{ 12, 12, 12, 12, 12, 12, 16, 20, 24, 28, 32, 40, 48, 56, 64, 76, 90, 2, 2, 2, 2, 2 }, /* 8000 */
};

static const uint8_t band_size_short[9][13] = {
{ 4, 4, 4, 4, 6,  8, 10, 12, 14, 18, 22, 30, 56, }, /* 44100 */
{ 4, 4, 4, 4, 6,  6, 10, 12, 14, 16, 20, 26, 66, }, /* 48000 */
{ 4, 4, 4, 4, 6,  8, 12, 16, 20, 26, 34, 42, 12, }, /* 32000 */
{ 4, 4, 4, 6, 6,  8, 10, 14, 18, 26, 32, 42, 18, }, /* 22050 */
{ 4, 4, 4, 6, 8, 10, 12, 14, 18, 24, 32, 44, 12, }, /* 24000 */
{ 4, 4, 4, 6, 8, 10, 12, 14, 18, 24, 30, 40, 18, }, /* 16000 */
{ 4, 4, 4, 6, 8, 10, 12, 14, 18, 24, 30, 40, 18, }, /* 11025 */
{ 4, 4, 4, 6, 8, 10, 12, 14, 18, 24, 30, 40, 18, }, /* 12000 */
{ 8, 8, 8, 12, 16, 20, 24, 28, 36, 2, 2, 2, 26, }, /* 8000 */
};

static const uint8_t mpa_pretab[2][22] = {
    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 3, 3, 3, 2, 0 },
};

/* table for alias reduction */
static const float ci_table[8] = {
    -0.6, -0.535, -0.33, -0.185, -0.095, -0.041, -0.0142, -0.0037,
};

/**
 * Convert region offsets to region sizes and truncate
 * size to big_values.
 */
static void ff_region_offset2size(GranuleDef *g){
    int i, k, j=0;
    g->region_size[2] = (576 / 2);
    for(i=0;i<3;i++) {
        k = FFMIN(g->region_size[i], g->big_values);
        g->region_size[i] = k - j;
        j = k;
    }
}

static void ff_init_short_region(MPADecodeContext *s, GranuleDef *g){
    if (g->block_type == 2) {
        g->region_size[0] = (36 / 2);
    } else {
        if (s->hdr.sample_rate_index <= 2)
            g->region_size[0] = (36 / 2);
        else if (s->hdr.sample_rate_index != 8)
            g->region_size[0] = (54 / 2);
        else
            g->region_size[0] = (108 / 2);
    }
    g->region_size[1] = (576 / 2);
}

static void ff_init_long_region(MPADecodeContext *s, GranuleDef *g, int ra1, int ra2){
    int l;
    g->region_size[0] =
        band_index_long[s->hdr.sample_rate_index][ra1 + 1] >> 1;
    /* should not overflow */
    l = FFMIN(ra1 + ra2 + 2, 22);
    g->region_size[1] =
        band_index_long[s->hdr.sample_rate_index][l] >> 1;
}

static void ff_compute_band_indexes(MPADecodeContext *s, GranuleDef *g){
    if (g->block_type == 2) {
        if (g->switch_point) {
            /* if switched mode, we handle the 36 first samples as
                long blocks.  For 8000Hz, we handle the 48 first
                exponents as long blocks (XXX: check this!) */
            if (s->hdr.sample_rate_index <= 2)
                g->long_end = 8;
            else if (s->hdr.sample_rate_index != 8)
                g->long_end = 6;
            else
                g->long_end = 4; /* 8000 Hz */

            g->short_start = 2 + (s->hdr.sample_rate_index != 8);
        } else {
            g->long_end = 0;
            g->short_start = 0;
        }
    } else {
        g->short_start = 13;
        g->long_end = 22;
    }
}

/* layer 1 unscaling */
/* n = number of bits of the mantissa minus 1 */
#define MUL64(a,b) ((int64_t)(a) * (int64_t)(b))
static inline int l1_unscale(int n, int mant, int scale_factor)
{
    int shift, mod;
    int64_t val;

    shift = scale_factor_modshift[scale_factor];
    mod = shift & 3;
    shift >>= 2;
    val = MUL64(mant + (-1 << n) + 1, scale_factor_mult[n-1][mod]);
    shift += n;
    /* NOTE: at this point, 1 <= shift >= 21 + 15 */
    return (int)((val + (1LL << (shift - 1))) >> shift);
}

static inline int l2_unscale_group(int steps, int mant, int scale_factor)
{
    int shift, mod, val;

    shift = scale_factor_modshift[scale_factor];
    mod = shift & 3;
    shift >>= 2;

    val = (mant - (steps >> 1)) * scale_factor_mult2[steps >> 2][mod];
    /* NOTE: at this point, 0 <= shift <= 21 */
    if (shift > 0)
        val = (val + (1 << (shift - 1))) >> shift;
    return val;
}

/* compute value^(4/3) * 2^(exponent/4). It normalized to FRAC_BITS */
static inline int l3_unscale(int value, int exponent)
{
    unsigned int m;
    int e;

    e = table_4_3_exp  [4*value + (exponent&3)];
    m = table_4_3_value[4*value + (exponent&3)];
    e -= (exponent >> 2);
    if (e > 31)
        return 0;
    m = (m + (1 << (e-1))) >> e;
    return m;
}

static const float
C1  = -5.0000000000e-01,
C2  =  4.1666667908e-02, /* 0x3d2aaaab */
C3  = -1.3888889225e-03, /* 0xbab60b61 */
C4  =  2.4801587642e-05, /* 0x37d00d01 */
C5  = -2.7557314297e-07, /* 0xb493f27c */
C6  =  2.0875723372e-09, /* 0x310f74f6 */
C7  = -1.1359647598e-11; /* 0xad47d74e */

// Differs from libc cosf on [0, pi/2] at only the point 0x3fc460f8 (~1.5342093) by 0.0000001005828f
static inline float fast_cosf(float x)
{
    float z,r;
    z  = x*x;
    r  = z*(C2+z*(C3+z*(C4+z*(C5+z*(C6+z*C7)))));
    return 1.0f + (z*r + C1*z);
}

int mp3_decode_init(MPADecodeContext *s)
{
    static int init=0;
    int i, j, k;
    //ff_mpegaudiodec_init_mmx(s);

    if (!init) {
        ff_mpa_synth_init();

        /* scale factors table for layer 1/2 */
        for(i=0;i<64;i++) {
            int shift, mod;
            /* 1.0 (i = 3) is normalized to 2 ^ FRAC_BITS */
            shift = (i / 3);
            mod = i % 3;
            scale_factor_modshift[i] = mod | (shift << 2);
        }

        /* scale factor multiply for layer 1 */
        for(i=0;i<15;i++) {
            int n, norm;
            n = i + 2;
            norm = ((1UL << n) * FRAC_ONE) / ((1 << n) - 1);
            scale_factor_mult[i][0] = (norm * 1.0f          * 2.0f);
            scale_factor_mult[i][1] = (norm * 0.7937005259f * 2.0f);
            scale_factor_mult[i][2] = (norm * 0.6299605249f * 2.0f);
        }

        for(i=0;i<9;i++) {
            k = 0;
            for(j=0;j<22;j++) {
                band_index_long[i][j] = k;
                k += band_size_long[i][j];
            }
            band_index_long[i][22] = k;
        }

        /* compute n ^ (4/3) and store it in mantissa/exp format */
        //int_pow_init();
        mpegaudio_tableinit();

        for (i = 0; i < 4; i++)
            if (ff_mpa_quant_bits[i] < 0)
                for (j = 0; j < (1<<(-ff_mpa_quant_bits[i]+1)); j++) {
                    int val1, val2, val3, steps;
                    int val = j;
                    steps  = ff_mpa_quant_steps[i];
                    val1 = val % steps;
                    val /= steps;
                    val2 = val % steps;
                    val3 = val / steps;
                    division_tabs[i][j] = val1 + (val2 << 4) + (val3 << 8);
                }

        for(i=0;i<8;i++) {
            float ci, cs, ca;
            ci = ci_table[i];
            cs = 1.0f / sqrtf(1.0f + ci * ci);
            ca = cs * ci;
            csa_table_float[i][0] = cs;
            csa_table_float[i][1] = ca;
            csa_table_float[i][2] = ca + cs;
            csa_table_float[i][3] = ca - cs;
        }

        /* compute mdct windows */
        for(i=0;i<36;i++) {
            for(j=0; j<4; j++){
                float d;

                if(j==2 && i%3 != 1)
                    continue;

                d= fast_cosf((float)M_PI/2.0f - ((float)M_PI * (i + 0.5f) / 36.0f));
                if(j==1){
                    if     (i>=30) d= 0;
                    else if(i>=24) d= fast_cosf((float)M_PI/2.0f - ((float)M_PI * (i - 18 + 0.5f) / 12.0f));
                    else if(i>=18) d= 1;
                }else if(j==3){
                    if     (i<  6) d= 0;
                    else if(i< 12) d= fast_cosf((float)M_PI/2.0f - ((float)M_PI * (i -  6 + 0.5f) / 12.0f));
                    else if(i< 18) d= 1;
                }
                //merge last stage of imdct into the window coefficients
                if (i > 8) d = -d;

                if(j==2)
                    mdct_win[j][i/3] = (d / (1<<5));
                else
                    mdct_win[j][i  ] = (d / (1<<5));
            }
        }

        /* NOTE: we do frequency inversion adter the MDCT by changing
           the sign of the right window coefs */
        for(j=0;j<4;j++) {
            for(i=0;i<36;i+=2) {
                mdct_win[j + 4][i] = mdct_win[j][i];
                mdct_win[j + 4][i + 1] = -mdct_win[j][i + 1];
            }
        }

        init = 1;
    }

    return 0;
}

#define C03 (0.86602540378443864676f)
#define C00 (0.51763809020504152469f/2) //0
#define C01 (0.70710678118654752439f)   //1
#define C02 (1.93185165257813657349f/2) //2

static const float imdct6_twiddle[6] = {
    0.5043145, 0.5411961, 0.6302362, 0.8213399, 1.3065630, 3.8306508
};

/* 12 points IMDCT. We compute it "by hand" by factorizing obvious cases. */
static void imdct12(float *out, float *in)
{
    float in0, in1, in2, in3, in4, in5, t1, t2;

    in0= in[0*3];
    in1= in[1*3] + in[0*3];
    in2= in[2*3] + in[1*3];
    in3= in[3*3] + in[2*3];
    in4= in[4*3] + in[3*3];
    in5= in[5*3] + in[4*3];
    in5 += in3;
    in3 += in1;
    in3 += in3;

    in2= (in2*C03);
    in3= (in3*C03);

    t1 = (in0 - in4);
    t2 = (in1 - in5)*C01;

    out[ 7]=
    out[10]= imdct6_twiddle[1]*(t1 + t2);
    out[ 1]=
    out[ 4]= imdct6_twiddle[4]*(t1 - t2);

    in0 += (in4*0.5f);
    in4 = in0 + in2;
    in5 += 2*in1;
    in1 = (in5 + in3)*C00;
    out[ 8]=
    out[ 9]= imdct6_twiddle[0]*(in4 + in1);
    out[ 2]=
    out[ 3]= imdct6_twiddle[5]*(in4 - in1);

    in0 -= in2;
    in5 = (in5 - in3)*C02;
    out[ 0]=
    out[ 5]= imdct6_twiddle[3]*(in0 - in5);
    out[ 6]=
    out[11]= imdct6_twiddle[2]*(in0 + in5);
}

/* return the number of decoded frames */
static int mp_decode_layer1(MPADecodeContext *s)
{
    int bound, i, v, n, ch, j, mant;
    uint8_t allocation[MPA_MAX_CHANNELS][SBLIMIT];
    uint8_t scale_factors[MPA_MAX_CHANNELS][SBLIMIT];

    if (s->hdr.mode == MPA_JSTEREO)
        bound = (s->hdr.mode_ext + 1) * 4;
    else
        bound = SBLIMIT;

    /* allocation bits */
    for(i=0;i<bound;i++) {
        for(ch=0;ch<s->nb_channels;ch++) {
            allocation[ch][i] = get_bits(&s->gb, 4);
        }
    }
    for(i=bound;i<SBLIMIT;i++) {
        allocation[0][i] = get_bits(&s->gb, 4);
    }

    /* scale factors */
    for(i=0;i<bound;i++) {
        for(ch=0;ch<s->nb_channels;ch++) {
            if (allocation[ch][i])
                scale_factors[ch][i] = get_bits(&s->gb, 6);
        }
    }
    for(i=bound;i<SBLIMIT;i++) {
        if (allocation[0][i]) {
            scale_factors[0][i] = get_bits(&s->gb, 6);
            scale_factors[1][i] = get_bits(&s->gb, 6);
        }
    }

    /* compute samples */
    for(j=0;j<12;j++) {
        for(i=0;i<bound;i++) {
            for(ch=0;ch<s->nb_channels;ch++) {
                n = allocation[ch][i];
                if (n) {
                    mant = get_bits(&s->gb, n + 1);
                    v = l1_unscale(n, mant, scale_factors[ch][i]);
                } else {
                    v = 0;
                }
                s->sb_samples[ch][j][i] = v;
            }
        }
        for(i=bound;i<SBLIMIT;i++) {
            n = allocation[0][i];
            if (n) {
                mant = get_bits(&s->gb, n + 1);
                v = l1_unscale(n, mant, scale_factors[0][i]);
                s->sb_samples[0][j][i] = v;
                v = l1_unscale(n, mant, scale_factors[1][i]);
                s->sb_samples[1][j][i] = v;
            } else {
                s->sb_samples[0][j][i] = 0;
                s->sb_samples[1][j][i] = 0;
            }
        }
    }
    return 12;
}

/* encoding tables which give the quantization index. Note how it is
   possible to store them efficiently ! */
static const unsigned char alloc_table_1[] = {
 4,  0,  2,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16,
 4,  0,  2,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16,
 4,  0,  2,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16,
 4,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 16,
 4,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 16,
 4,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 16,
 4,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 16,
 4,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 16,
 4,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 16,
 4,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 16,
 4,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 16,
 3,  0,  1,  2,  3,  4,  5, 16,
 3,  0,  1,  2,  3,  4,  5, 16,
 3,  0,  1,  2,  3,  4,  5, 16,
 3,  0,  1,  2,  3,  4,  5, 16,
 3,  0,  1,  2,  3,  4,  5, 16,
 3,  0,  1,  2,  3,  4,  5, 16,
 3,  0,  1,  2,  3,  4,  5, 16,
 3,  0,  1,  2,  3,  4,  5, 16,
 3,  0,  1,  2,  3,  4,  5, 16,
 3,  0,  1,  2,  3,  4,  5, 16,
 3,  0,  1,  2,  3,  4,  5, 16,
 3,  0,  1,  2,  3,  4,  5, 16,
 2,  0,  1, 16,
 2,  0,  1, 16,
 2,  0,  1, 16,
 2,  0,  1, 16,
 2,  0,  1, 16,
 2,  0,  1, 16,
 2,  0,  1, 16,
};

static const unsigned char alloc_table_3[] = {
 4,  0,  1,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
 4,  0,  1,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
 3,  0,  1,  3,  4,  5,  6,  7,
 3,  0,  1,  3,  4,  5,  6,  7,
 3,  0,  1,  3,  4,  5,  6,  7,
 3,  0,  1,  3,  4,  5,  6,  7,
 3,  0,  1,  3,  4,  5,  6,  7,
 3,  0,  1,  3,  4,  5,  6,  7,
 3,  0,  1,  3,  4,  5,  6,  7,
 3,  0,  1,  3,  4,  5,  6,  7,
 3,  0,  1,  3,  4,  5,  6,  7,
 3,  0,  1,  3,  4,  5,  6,  7,
};

static const unsigned char alloc_table_4[] = {
 4,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
 4,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
 4,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
 4,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
 3,  0,  1,  3,  4,  5,  6,  7,
 3,  0,  1,  3,  4,  5,  6,  7,
 3,  0,  1,  3,  4,  5,  6,  7,
 3,  0,  1,  3,  4,  5,  6,  7,
 3,  0,  1,  3,  4,  5,  6,  7,
 3,  0,  1,  3,  4,  5,  6,  7,
 3,  0,  1,  3,  4,  5,  6,  7,
 2,  0,  1,  3,
 2,  0,  1,  3,
 2,  0,  1,  3,
 2,  0,  1,  3,
 2,  0,  1,  3,
 2,  0,  1,  3,
 2,  0,  1,  3,
 2,  0,  1,  3,
 2,  0,  1,  3,
 2,  0,  1,  3,
 2,  0,  1,  3,
 2,  0,  1,  3,
 2,  0,  1,  3,
 2,  0,  1,  3,
 2,  0,  1,  3,
 2,  0,  1,  3,
 2,  0,  1,  3,
 2,  0,  1,  3,
 2,  0,  1,  3,
};

static const unsigned char * const ff_mpa_alloc_tables[5] =
{ alloc_table_1, alloc_table_1, alloc_table_3, alloc_table_3, alloc_table_4, };

static int mp_decode_layer2(MPADecodeContext *s)
{
    int sblimit; /* number of used subbands */
    const unsigned char *alloc_table;
    int table, bit_alloc_bits, i, j, ch, bound, v;
    unsigned char bit_alloc[MPA_MAX_CHANNELS][SBLIMIT];
    unsigned char scale_code[MPA_MAX_CHANNELS][SBLIMIT];
    unsigned char scale_factors[MPA_MAX_CHANNELS][SBLIMIT][3], *sf;
    int scale, qindex, bits, steps, k, l, m, b;

    /* select decoding table */
    table = ff_mpa_l2_select_table(s->hdr.bit_rate / 1000, s->nb_channels, s->hdr.sample_rate, s->hdr.lsf);
    sblimit = ff_mpa_sblimit_table[table];
    alloc_table = ff_mpa_alloc_tables[table];

    if (s->hdr.mode == MPA_JSTEREO)
        bound = (s->hdr.mode_ext + 1) * 4;
    else
        bound = sblimit;

    /* sanity check */
    if( bound > sblimit ) bound = sblimit;

    /* parse bit allocation */
    j = 0;
    for(i=0;i<bound;i++) {
        bit_alloc_bits = alloc_table[j];
        for(ch=0;ch<s->nb_channels;ch++) {
            bit_alloc[ch][i] = get_bits(&s->gb, bit_alloc_bits);
        }
        j += 1 << bit_alloc_bits;
    }
    for(i=bound;i<sblimit;i++) {
        bit_alloc_bits = alloc_table[j];
        v = get_bits(&s->gb, bit_alloc_bits);
        bit_alloc[0][i] = v;
        bit_alloc[1][i] = v;
        j += 1 << bit_alloc_bits;
    }

    /* scale codes */
    for(i=0;i<sblimit;i++) {
        for(ch=0;ch<s->nb_channels;ch++) {
            if (bit_alloc[ch][i])
                scale_code[ch][i] = get_bits(&s->gb, 2);
        }
    }

    /* scale factors */
    for(i=0;i<sblimit;i++) {
        for(ch=0;ch<s->nb_channels;ch++) {
            if (bit_alloc[ch][i]) {
                sf = scale_factors[ch][i];
                switch(scale_code[ch][i]) {
                default:
                case 0:
                    sf[0] = get_bits(&s->gb, 6);
                    sf[1] = get_bits(&s->gb, 6);
                    sf[2] = get_bits(&s->gb, 6);
                    break;
                case 2:
                    sf[0] = get_bits(&s->gb, 6);
                    sf[1] = sf[0];
                    sf[2] = sf[0];
                    break;
                case 1:
                    sf[0] = get_bits(&s->gb, 6);
                    sf[2] = get_bits(&s->gb, 6);
                    sf[1] = sf[0];
                    break;
                case 3:
                    sf[0] = get_bits(&s->gb, 6);
                    sf[2] = get_bits(&s->gb, 6);
                    sf[1] = sf[2];
                    break;
                }
            }
        }
    }

    /* samples */
    for(k=0;k<3;k++) {
        for(l=0;l<12;l+=3) {
            j = 0;
            for(i=0;i<bound;i++) {
                bit_alloc_bits = alloc_table[j];
                for(ch=0;ch<s->nb_channels;ch++) {
                    b = bit_alloc[ch][i];
                    if (b) {
                        scale = scale_factors[ch][i][k];
                        qindex = alloc_table[j+b];
                        bits = ff_mpa_quant_bits[qindex];
                        if (bits < 0) {
                            int v2;
                            /* 3 values at the same time */
                            v = get_bits(&s->gb, -bits);
                            v2 = division_tabs[qindex][v];
                            steps  = ff_mpa_quant_steps[qindex];

                            s->sb_samples[ch][k * 12 + l + 0][i] =
                                l2_unscale_group(steps, v2        & 15, scale);
                            s->sb_samples[ch][k * 12 + l + 1][i] =
                                l2_unscale_group(steps, (v2 >> 4) & 15, scale);
                            s->sb_samples[ch][k * 12 + l + 2][i] =
                                l2_unscale_group(steps,  v2 >> 8      , scale);
                        } else {
                            for(m=0;m<3;m++) {
                                v = get_bits(&s->gb, bits);
                                v = l1_unscale(bits - 1, v, scale);
                                s->sb_samples[ch][k * 12 + l + m][i] = v;
                            }
                        }
                    } else {
                        s->sb_samples[ch][k * 12 + l + 0][i] = 0;
                        s->sb_samples[ch][k * 12 + l + 1][i] = 0;
                        s->sb_samples[ch][k * 12 + l + 2][i] = 0;
                    }
                }
                /* next subband in alloc table */
                j += 1 << bit_alloc_bits;
            }
            /* XXX: find a way to avoid this duplication of code */
            for(i=bound;i<sblimit;i++) {
                bit_alloc_bits = alloc_table[j];
                b = bit_alloc[0][i];
                if (b) {
                    int mant, scale0, scale1;
                    scale0 = scale_factors[0][i][k];
                    scale1 = scale_factors[1][i][k];
                    qindex = alloc_table[j+b];
                    bits = ff_mpa_quant_bits[qindex];
                    if (bits < 0) {
                        /* 3 values at the same time */
                        v = get_bits(&s->gb, -bits);
                        steps = ff_mpa_quant_steps[qindex];
                        mant = v % steps;
                        v = v / steps;
                        s->sb_samples[0][k * 12 + l + 0][i] =
                            l2_unscale_group(steps, mant, scale0);
                        s->sb_samples[1][k * 12 + l + 0][i] =
                            l2_unscale_group(steps, mant, scale1);
                        mant = v % steps;
                        v = v / steps;
                        s->sb_samples[0][k * 12 + l + 1][i] =
                            l2_unscale_group(steps, mant, scale0);
                        s->sb_samples[1][k * 12 + l + 1][i] =
                            l2_unscale_group(steps, mant, scale1);
                        s->sb_samples[0][k * 12 + l + 2][i] =
                            l2_unscale_group(steps, v, scale0);
                        s->sb_samples[1][k * 12 + l + 2][i] =
                            l2_unscale_group(steps, v, scale1);
                    } else {
                        for(m=0;m<3;m++) {
                            mant = get_bits(&s->gb, bits);
                            s->sb_samples[0][k * 12 + l + m][i] =
                                l1_unscale(bits - 1, mant, scale0);
                            s->sb_samples[1][k * 12 + l + m][i] =
                                l1_unscale(bits - 1, mant, scale1);
                        }
                    }
                } else {
                    s->sb_samples[0][k * 12 + l + 0][i] = 0;
                    s->sb_samples[0][k * 12 + l + 1][i] = 0;
                    s->sb_samples[0][k * 12 + l + 2][i] = 0;
                    s->sb_samples[1][k * 12 + l + 0][i] = 0;
                    s->sb_samples[1][k * 12 + l + 1][i] = 0;
                    s->sb_samples[1][k * 12 + l + 2][i] = 0;
                }
                /* next subband in alloc table */
                j += 1 << bit_alloc_bits;
            }
            /* fill remaining samples to zero */
            for(i=sblimit;i<SBLIMIT;i++) {
                for(ch=0;ch<s->nb_channels;ch++) {
                    s->sb_samples[ch][k * 12 + l + 0][i] = 0;
                    s->sb_samples[ch][k * 12 + l + 1][i] = 0;
                    s->sb_samples[ch][k * 12 + l + 2][i] = 0;
                }
            }
        }
    }
    return 3 * 12;
}

#define SPLIT(dst,sf,n)\
    if(n==3){\
        int m= (sf*171)>>9;\
        dst= sf - 3*m;\
        sf=m;\
    }else if(n==4){\
        dst= sf&3;\
        sf>>=2;\
    }else if(n==5){\
        int m= (sf*205)>>10;\
        dst= sf - 5*m;\
        sf=m;\
    }else if(n==6){\
        int m= (sf*171)>>10;\
        dst= sf - 6*m;\
        sf=m;\
    }else{\
        dst=0;\
    }

static __attribute__((always_inline)) inline void lsf_sf_expand(int *slen, int sf, int n1, int n2, int n3)
{
    SPLIT(slen[3], sf, n3)
    SPLIT(slen[2], sf, n2)
    SPLIT(slen[1], sf, n1)
    slen[0] = sf;
}

static void exponents_from_scale_factors(MPADecodeContext *s, GranuleDef *g, int16_t *exponents)
{
    const uint8_t *bstab, *pretab;
    int len, i, j, k, l, v0, shift, gain, gains[3];
    int16_t *exp_ptr;

    exp_ptr = exponents;
    gain = g->global_gain - 210;
    shift = g->scalefac_scale + 1;

    bstab = band_size_long[s->hdr.sample_rate_index];
    pretab = mpa_pretab[g->preflag];
    for(i=0;i<g->long_end;i++) {
        v0 = gain - ((g->scale_factors[i] + pretab[i]) << shift) + 400;
        len = bstab[i];
        for(j=len;j>0;j--)
            *exp_ptr++ = v0;
    }

    if (g->short_start < 13) {
        bstab = band_size_short[s->hdr.sample_rate_index];
        gains[0] = gain - (g->subblock_gain[0] << 3);
        gains[1] = gain - (g->subblock_gain[1] << 3);
        gains[2] = gain - (g->subblock_gain[2] << 3);
        k = g->long_end;
        for(i=g->short_start;i<13;i++) {
            len = bstab[i];
            for(l=0;l<3;l++) {
                v0 = gains[l] - (g->scale_factors[k++] << shift) + 400;
                for(j=len;j>0;j--)
                *exp_ptr++ = v0;
            }
        }
    }
}

/* handle n = 0 too */
static inline int get_bitsz(GetBitContext *s, int n)
{
    if (n == 0) return 0;
    else return get_bits(s, n);
}

static void switch_buffer(MPADecodeContext *s, int *pos, int *end_pos, int *end_pos2){
    if(s->in_gb.buffer && *pos >= s->gb.buffer_size*8){
        s->gb= s->in_gb;
        s->in_gb.buffer=NULL;
        skip_bits(&s->gb, *pos - *end_pos);
        *end_pos2=
        *end_pos= *end_pos2 + get_bits_count(&s->gb) - *pos;
        *pos= get_bits_count(&s->gb);
    }
}

/* Following is a optimized code for
            float v = *src
            if(get_bits1(&s->gb))
                v = -v;
            *dst = v;
*/

static inline float READ_FLIP_SIGN(float *src, uint32_t bit) {
    _f32 tmp;
    tmp.f = *src; tmp.i ^= (bit<<31);
    return tmp.f;
}

static inline float READ43_FLIP_SIGN(uint32_t x, uint32_t bit) {
    _f32 tmp;
    tmp.f = pow4_3_tab[x&15]; tmp.i ^= (bit<<31);
    return tmp.f;
}

/**
 * parses a vlc code, faster then get_vlc()
 * @param bits is the number of bits which will be read at once, must be
 *             identical to nb_bits in init_vlc()
 * @param max_depth is the number of times bits bits must be read to completely
 *                  read the longest vlc code
 *                  = (max_vlc_length + bits - 1) / bits
 */
int get_vlc2(GetBitContext *s, VLC_TYPE (*table)[2], int bits, int max_depth)
{
    unsigned int index, re_index = s->index;
    unsigned int re_cache = AV_RB32(s->buffer + (re_index >> 3)) << (re_index & 7);
    int n, nb_bits, code;
    index = NEG_USR32(re_cache, bits);
    code  = table[index][0];
    n     = table[index][1];
    if (n < 0) {
        re_index += bits;
        re_cache = AV_RB32(s->buffer + (re_index >> 3)) << (re_index & 7);
        nb_bits = -n;
        index = NEG_USR32(re_cache, nb_bits) + code;
        code  = table[index][0];
        n     = table[index][1];
        if (n < 0) {
            re_index += nb_bits;
            re_cache = AV_RB32(s->buffer + (re_index >> 3)) << (re_index & 7);
            nb_bits = -n;
            index = NEG_USR32(re_cache, nb_bits) + code;
            code  = table[index][0];
            n     = table[index][1];
        }
    }
    re_index += n;
    s->index = re_index;
    return code;
}

static int huffman_decode(MPADecodeContext *s, GranuleDef *g,
                          int16_t *exponents, int end_pos2)
{
    int s_index;
    int i, n;
    int last_pos, bits_left;
    const VLC *vlc;
    int end_pos= FFMIN(end_pos2, s->gb.buffer_size*8);

    /* low frequencies (called big values) */
    s_index = 0;
    for(i=0;i<3;i++) {
        int j, k, l, linbits;
        j = g->region_size[i];
        if (j == 0)
            continue;
        /* select vlc table */
        k = g->table_select[i];
        l = mpa_huff_data[k][0];
        linbits = mpa_huff_data[k][1];
        vlc = &mp3_huff_vlc[l];
        //av_log(AV_LOG_INFO, "frame %lu: table: %u, linbits: %u, huffman table size: %u\n", s->frame_number, l, linbits, vlc->bits);
        if(!l){
            for (n = 0; n < 2*j; n++) {
                g->sb_hybrid[s_index + n] = 0;
            }
            s_index += 2*j;
            continue;
        }

        /* read huffcode and compute each couple */
        for(;j>0;j--) {
            int exponent, x, y;
            int v;
            int pos= get_bits_count(&s->gb);

            if (pos >= end_pos){
                //av_log(AV_LOG_ERROR, "pos: %d %d %d %d\n", pos>>3, end_pos>>3, end_pos2>>3, s_index);
                switch_buffer(s, &pos, &end_pos, &end_pos2);
                //av_log(AV_LOG_ERROR, "new pos: %d %d\n", pos>>3, end_pos>>3);
                if(pos >= end_pos)
                    break;
            }
            y = get_vlc2(&s->gb, vlc->table, vlc->bits, 3);

            if(!y){
                g->sb_hybrid[s_index  ] =
                g->sb_hybrid[s_index+1] = 0;
                s_index += 2;
                continue;
            }
            exponent= exponents[s_index];

            if(y&16){
                x = y >> 5;
                y = y & 0x0f;
                if (x < 15){
                    g->sb_hybrid[s_index] = exp_table_float[exponent]*READ43_FLIP_SIGN(x, get_bits1(&s->gb));
                }else{
                    x += get_bitsz(&s->gb, linbits);
                    v = l3_unscale(x, exponent);
                    if (get_bits1(&s->gb))
                        v = -v;
                    g->sb_hybrid[s_index] = v;
                }
                if (y < 15){
                    g->sb_hybrid[s_index+1] = exp_table_float[exponent]*READ43_FLIP_SIGN(y, get_bits1(&s->gb));
                }else{
                    y += get_bitsz(&s->gb, linbits);
                    v = l3_unscale(y, exponent);
                    if (get_bits1(&s->gb))
                        v = -v;
                    g->sb_hybrid[s_index+1] = v;
                }
            }else{
                x = y >> 5;
                y = y & 0x0f;
                x += y;
                if (x < 15){
                    g->sb_hybrid[s_index+!!y] = exp_table_float[exponent]*READ43_FLIP_SIGN(x, get_bits1(&s->gb));
                }else{
                    x += get_bitsz(&s->gb, linbits);
                    v = l3_unscale(x, exponent);
                    if (get_bits1(&s->gb))
                        v = -v;
                    g->sb_hybrid[s_index+!!y] = v;
                }
                g->sb_hybrid[s_index+ !y] = 0;
            }
            s_index+=2;
        }
    }

    /* high frequencies */
    last_pos=0;
    while (s_index <= 572) {
        int pos, codeidx, code;
        pos = get_bits_count(&s->gb);
        if (pos >= end_pos) {
            if (pos > end_pos2 && last_pos){
                /* some encoders generate an incorrect size for this
                   part. We must go back into the data */
                s_index -= 4;
                skip_bits(&s->gb, last_pos - pos);
                av_log(AV_LOG_INFO, "overread, skip %d enddists: %d %d\n", last_pos - pos, end_pos-pos, end_pos2-pos);
                break;
            }
            switch_buffer(s, &pos, &end_pos, &end_pos2);
            if(pos >= end_pos)
                break;
        }
        last_pos= pos;

        if (g->count1table_select) {
            codeidx = show_bits(&s->gb, 4);
            skip_bits(&s->gb, 4);
            code = huff_quad_vlc_table2[codeidx][0];
        } else {
            codeidx = show_bits(&s->gb, 7);
            /*if (codeidx > 31) {
                skip_bits(&s->gb, 1);
                code = 0;
            } else {*/
                skip_bits(&s->gb, huff_quad_vlc_table1[codeidx][1]);
                code = huff_quad_vlc_table1[codeidx][0];
            //}
        }
        g->sb_hybrid[s_index+0]=
        g->sb_hybrid[s_index+1]=
        g->sb_hybrid[s_index+2]=
        g->sb_hybrid[s_index+3]= 0;
        while(code){
            static const int idxtab[16]={3,3,2,2,1,1,1,1,0,0,0,0,0,0,0,0};
            int pos= s_index+idxtab[code];
            code ^= 8>>idxtab[code];
            g->sb_hybrid[pos] = READ_FLIP_SIGN((exp_table_float+exponents[pos]), get_bits1(&s->gb));
        }
        s_index+=4;
    }
    /* skip extension bits */
    bits_left = end_pos2 - get_bits_count(&s->gb);
    if (bits_left < 0) {
        av_log(AV_LOG_ERROR, "bits_left=%d\n", bits_left);
        s_index=0;
    }
    /*if (bits_left)
        av_log(AV_LOG_ERROR, "bits_left=%d\n", bits_left);*/
    skip_bits(&s->gb, bits_left);
    for (n = 0; n < (576 - s_index); n++) {
        g->sb_hybrid[s_index + n] = 0;
    }
    i= get_bits_count(&s->gb);
    switch_buffer(s, &i, &end_pos, &end_pos2);
    return 0;
}

/* Reorder short blocks from bitstream order to interleaved order. It
   would be faster to do it in parsing, but the code would be far more
   complicated */
static void reorder_block(MPADecodeContext *s, GranuleDef *g)
{
    int i, j, len;
    float *ptr, *dst, *ptr1;
    float tmp[576];

    if (g->block_type != 2)
        return;

    if (g->switch_point) {
        if (s->hdr.sample_rate_index != 8) {
            ptr = g->sb_hybrid + 36;
        } else {
            ptr = g->sb_hybrid + 48;
        }
    } else {
        ptr = g->sb_hybrid;
    }

    for(i=g->short_start;i<13;i++) {
        len = band_size_short[s->hdr.sample_rate_index][i];
        ptr1 = ptr;
        dst = tmp;
        for(j=len;j>0;j--) {
            *dst++ = ptr[0*len];
            *dst++ = ptr[1*len];
            *dst++ = ptr[2*len];
            ptr++;
        }
        ptr+=2*len;
        memcpy(ptr1, tmp, len * 3 * sizeof(float));
    }
}
#define ISQRT2 0.70710678118654752440f

static void compute_stereo(MPADecodeContext *s,
                           GranuleDef *g0, GranuleDef *g1)
{
    int i, j, k, l;
    int sf_max, sf, len, non_zero_found;
    float (*is_tab)[16], *tab0, *tab1, tmp0, tmp1, v1, v2;
    int non_zero_found_short[3];

    /* intensity stereo */
    if (s->hdr.mode_ext & MODE_EXT_I_STEREO) {
        if (!s->hdr.lsf) {
            is_tab = is_table;
            sf_max = 7;
        } else {
            is_tab = ((g1->scalefac_compress & 1) ? is_table_lsf1 : is_table_lsf0);
            sf_max = 16;
        }

        tab0 = g0->sb_hybrid + 576;
        tab1 = g1->sb_hybrid + 576;

        non_zero_found_short[0] = 0;
        non_zero_found_short[1] = 0;
        non_zero_found_short[2] = 0;
        k = (13 - g1->short_start) * 3 + g1->long_end - 3;
        for(i = 12;i >= g1->short_start;i--) {
            /* for last band, use previous scale factor */
            if (i != 11)
                k -= 3;
            len = band_size_short[s->hdr.sample_rate_index][i];
            for(l=2;l>=0;l--) {
                tab0 -= len;
                tab1 -= len;
                if (!non_zero_found_short[l]) {
                    /* test if non zero band. if so, stop doing i-stereo */
                    for(j=0;j<len;j++) {
                        if (tab1[j] != 0) {
                            non_zero_found_short[l] = 1;
                            goto found1;
                        }
                    }
                    sf = g1->scale_factors[k + l];
                    if (sf >= sf_max)
                        goto found1;

                    v1 = is_tab[0][sf];
                    v2 = is_tab[1][sf];
                    for(j=0;j<len;j++) {
                        tmp0 = tab0[j];
                        tab0[j] = (tmp0*v1);
                        tab1[j] = (tmp0*v2);
                    }
                } else {
                found1:
                    if (s->hdr.mode_ext & MODE_EXT_MS_STEREO) {
                        /* lower part of the spectrum : do ms stereo if enabled */
                        for(j=0;j<len;j++) {
                            tmp0 = tab0[j];
                            tmp1 = tab1[j];
                            tab0[j] = ((tmp0 + tmp1) * ISQRT2);
                            tab1[j] = ((tmp0 - tmp1) * ISQRT2);
                        }
                    }
                }
            }
        }

        non_zero_found = non_zero_found_short[0] |
            non_zero_found_short[1] |
            non_zero_found_short[2];

        for(i = g1->long_end - 1;i >= 0;i--) {
            len = band_size_long[s->hdr.sample_rate_index][i];
            tab0 -= len;
            tab1 -= len;
            /* test if non zero band. if so, stop doing i-stereo */
            if (!non_zero_found) {
                for(j=0;j<len;j++) {
                    if (tab1[j] != 0) {
                        non_zero_found = 1;
                        goto found2;
                    }
                }
                /* for last band, use previous scale factor */
                k = (i == 21) ? 20 : i;
                sf = g1->scale_factors[k];
                if (sf >= sf_max)
                    goto found2;
                v1 = is_tab[0][sf];
                v2 = is_tab[1][sf];
                for(j=0;j<len;j++) {
                    tmp0 = tab0[j];
                    tab0[j] = (tmp0*v1);
                    tab1[j] = (tmp0*v2);
                }
            } else {
            found2:
                if (s->hdr.mode_ext & MODE_EXT_MS_STEREO) {
                    /* lower part of the spectrum : do ms stereo if enabled */
                    for(j=0;j<len;j++) {
                        tmp0 = tab0[j];
                        tmp1 = tab1[j];
                        tab0[j] = ((tmp0 + tmp1) * ISQRT2);
                        tab1[j] = ((tmp0 - tmp1) * ISQRT2);
                    }
                }
            }
        }
    } else if (s->hdr.mode_ext & MODE_EXT_MS_STEREO) {
        /* ms stereo ONLY */
        /* NOTE: the 1/sqrt(2) normalization factor is included in the global gain */
        //butterflies_float_sse(g0->sb_hybrid, g1->sb_hybrid, 576);
        tab0 = g0->sb_hybrid;
        tab1 = g1->sb_hybrid;
        for(i=0;i<576;i++) {
            tmp0 = tab0[i];
            tmp1 = tab1[i];
            tab0[i] = tmp0 + tmp1;
            tab1[i] = tmp0 - tmp1;
        }
    }
}

static void imdct36_window(float *out, float *buf, float *in, float *win)
{
    int j;
    float t0, t1;

    for(j=0;j<4;j++) {
        t0 = in[    j];
        t1 = in[9+8-j];
        out[(9 + j)*SBLIMIT] =  (t1*win[9 + j]) + buf[9 + j];
        out[(8 - j)*SBLIMIT] =  (t1*win[8 - j]) + buf[8 - j];
        buf[9 + j] = (t0*win[18 + 9 + j]);
        buf[8 - j] = (t0*win[18 + 8 - j]);

        t0 = in[  8-j];
        t1 = in[9+  j];
        out[(9 + 8 - j)*SBLIMIT] =  (t1*win[9 + 8 - j]) + buf[9 + 8 - j];
        out[(        j)*SBLIMIT] =  (t1*win[        j]) + buf[        j];
        buf[9 + 8 - j] = (t0*win[18 + 9 + 8 - j]);
        buf[      + j] = (t0*win[18         + j]);
    }

    t0 = in[4];
    t1 = in[9+4];
    out[(9 + 4)*SBLIMIT] =  (t1*win[9 + 4]) + buf[9 + 4];
    out[(8 - 4)*SBLIMIT] =  (t1*win[8 - 4]) + buf[8 - 4];
    buf[9 + 4] = (t0*win[18 + 9 + 4]);
    buf[8 - 4] = (t0*win[18 + 8 - 4]);
}

static void compute_imdct(MPADecodeContext *s, GranuleDef *g, float *sb_samples, float *mdct_buf)
{
    DECLARE_ALIGNED(16, float, tmpbuf[18]);
    float *win, *win1, *out_ptr, *ptr, *buf, *ptr1;
    float out2[12];
    int i, j, mdct_long_end, sblimit;

    /* find last non zero block */
    ptr = g->sb_hybrid + 576;
    ptr1 = g->sb_hybrid + 2 * 18;
    while (ptr >= ptr1) {
        int32_t *p;
        ptr -= 6;
        p= (int32_t*)ptr;
        if(p[0] | p[1] | p[2] | p[3] | p[4] | p[5])
            break;
    }
    sblimit = ((ptr - g->sb_hybrid) / 18) + 1;

    if (g->block_type == 2) {
        /* XXX: check for 8000 Hz */
        if (g->switch_point)
            mdct_long_end = 2;
        else
            mdct_long_end = 0;
    } else {
        mdct_long_end = sblimit;
    }

    buf = mdct_buf;
    ptr = g->sb_hybrid;
    for(j=0;j<mdct_long_end;j++) {
        /* compute 18-point IMDCT */
        imdct36_sse(tmpbuf, ptr);

        /* apply window & overlap with previous buffer */
        out_ptr = sb_samples + j;
        /* select window */
        if (g->switch_point && j < 2)
            win1 = mdct_win[0];
        else
            win1 = mdct_win[g->block_type];
        /* select frequency inversion */
        win = win1 + ((4 * 36) & -(j & 1));
        imdct36_window(out_ptr, buf, tmpbuf, win);
        out_ptr += 18*SBLIMIT;
        ptr += 18;
        buf += 18;
    }
    for(j=mdct_long_end;j<sblimit;j++) {
        /* select frequency inversion */
        win = mdct_win[2] + ((4 * 36) & -(j & 1));
        out_ptr = sb_samples + j;

        for(i=0; i<6; i++){
            *out_ptr = buf[i];
            out_ptr += SBLIMIT;
        }
        imdct12(out2, ptr + 0);
        for(i=0;i<6;i++) {
            *out_ptr     = (out2[i    ]*win[i    ]) + buf[i + 6*1];
            buf[i + 6*2] = (out2[i + 6]*win[i + 6]);
            out_ptr += SBLIMIT;
        }
        imdct12(out2, ptr + 1);
        for(i=0;i<6;i++) {
            *out_ptr     = (out2[i    ]*win[i    ]) + buf[i + 6*2];
            buf[i + 6*0] = (out2[i + 6]*win[i + 6]);
            out_ptr += SBLIMIT;
        }
        imdct12(out2, ptr + 2);
        for(i=0;i<6;i++) {
            buf[i + 6*0] = (out2[i    ]*win[i    ]) + buf[i + 6*0];
            buf[i + 6*1] = (out2[i + 6]*win[i + 6]);
            buf[i + 6*2] = 0;
        }
        ptr += 18;
        buf += 18;
    }
    /* zero bands */
    for(j=sblimit;j<SBLIMIT;j++) {
        /* overlap */
        out_ptr = sb_samples + j;
        for(i=0;i<18;i++) {
            *out_ptr = buf[i];
            buf[i] = 0;
            out_ptr += SBLIMIT;
        }
        buf += 18;
    }
}

static void compute_antialias(MPADecodeContext *s, GranuleDef *g)
{
    float *ptr;
    int n, i;

    /* we antialias only "long" bands */
    if (g->block_type == 2) {
        if (!g->switch_point)
            return;
        /* XXX: check this for 8000Hz case */
        n = 1;
    } else {
        n = SBLIMIT - 1;
    }

    ptr = g->sb_hybrid + 18;
    for(i = n;i > 0;i--) {
        float tmp0, tmp1;
        float *csa = &csa_table_float[0][0];
#define FLOAT_AA(j)\
        tmp0= ptr[-1-j];\
        tmp1= ptr[   j];\
        ptr[-1-j] = tmp0 * csa[0+4*j] - tmp1 * csa[1+4*j];\
        ptr[   j] = tmp0 * csa[1+4*j] + tmp1 * csa[0+4*j];

        FLOAT_AA(0)
        FLOAT_AA(1)
        FLOAT_AA(2)
        FLOAT_AA(3)
        FLOAT_AA(4)
        FLOAT_AA(5)
        FLOAT_AA(6)
        FLOAT_AA(7)

        ptr += 18;
    }
}

/* main layer3 decoding function */
static int mp_decode_layer3(MPADecodeContext *s)
{
    int nb_granules, main_data_begin, private_bits;
    int gr, ch, blocksplit_flag, i, j, k, n, bits_pos;
    GranuleDef *g;
    int16_t exponents[576]; //FIXME try float

    /* read side info */
    if (s->hdr.lsf) {
        main_data_begin = get_bits(&s->gb, 8);
        private_bits = get_bits(&s->gb, s->nb_channels);
        nb_granules = 1;
    } else {
        main_data_begin = get_bits(&s->gb, 9);
        if (s->nb_channels == 2)
            private_bits = get_bits(&s->gb, 3);
        else
            private_bits = get_bits(&s->gb, 5);
        nb_granules = 2;
        for(ch=0;ch<s->nb_channels;ch++) {
            s->granules[ch][0].scfsi = 0;/* all scale factors are transmitted */
            s->granules[ch][1].scfsi = get_bits(&s->gb, 4);
        }
    }

    for(gr=0;gr<nb_granules;gr++) {
        for(ch=0;ch<s->nb_channels;ch++) {
            g = &s->granules[ch][gr];
            g->part2_3_length = get_bits(&s->gb, 12);
            g->big_values = get_bits(&s->gb, 9);
            if(g->big_values > 288){
                av_log(AV_LOG_ERROR, "big_values too big\n");
                return -1;
            }

            g->global_gain = get_bits(&s->gb, 8);
            /* if MS stereo only is selected, we precompute the
               1/sqrt(2) renormalization factor */
            if ((s->hdr.mode_ext & (MODE_EXT_MS_STEREO | MODE_EXT_I_STEREO)) == MODE_EXT_MS_STEREO)
                g->global_gain -= 2;
            if (s->hdr.lsf)
                g->scalefac_compress = get_bits(&s->gb, 9);
            else
                g->scalefac_compress = get_bits(&s->gb, 4);
            blocksplit_flag = get_bits1(&s->gb);
            if (blocksplit_flag) {
                g->block_type = get_bits(&s->gb, 2);
                if (g->block_type == 0){
                    av_log(AV_LOG_ERROR, "invalid block type\n");
                    return -1;
                }
                g->switch_point = get_bits1(&s->gb);
                for(i=0;i<2;i++)
                    g->table_select[i] = get_bits(&s->gb, 5);
                for(i=0;i<3;i++)
                    g->subblock_gain[i] = get_bits(&s->gb, 3);
                ff_init_short_region(s, g);
            } else {
                int region_address1, region_address2;
                g->block_type = 0;
                g->switch_point = 0;
                for(i=0;i<3;i++)
                    g->table_select[i] = get_bits(&s->gb, 5);
                /* compute huffman coded region sizes */
                region_address1 = get_bits(&s->gb, 4);
                region_address2 = get_bits(&s->gb, 3);
                ff_init_long_region(s, g, region_address1, region_address2);
            }
            ff_region_offset2size(g);
            ff_compute_band_indexes(s, g);

            g->preflag = 0;
            if (!s->hdr.lsf)
                g->preflag = get_bits1(&s->gb);
            g->scalefac_scale = get_bits1(&s->gb);
            g->count1table_select = get_bits1(&s->gb);
        }
    }

    if (!s->adu_mode) {
        const uint8_t *ptr = s->gb.buffer + (get_bits_count(&s->gb)>>3);
        /* now we get bits from the main_data_begin offset */
        //av_log(AV_LOG_DEBUG, "backstep:%d, lastbuf:%d\n", main_data_begin, s->last_buf_size);
        memcpy(s->last_buf + s->last_buf_size, ptr, EXTRABYTES);
        s->in_gb= s->gb;
        init_get_bits(&s->gb, s->last_buf, s->last_buf_size);
        skip_bits(&s->gb, 8*(s->last_buf_size - main_data_begin));
    }

    for(gr=0;gr<nb_granules;gr++) {
        for(ch=0;ch<s->nb_channels;ch++) {
            g = &s->granules[ch][gr];
            if(get_bits_count(&s->gb)<0){
                unsigned int l;
                /* av_log(AV_LOG_DEBUG, "mdb:%d, lastbuf:%d skipping granule %d\n",
                                            main_data_begin, s->last_buf_size, gr); */
                skip_bits(&s->gb, g->part2_3_length);
                for (l = 0; l < 576; l++) {
                    g->sb_hybrid[l] = 0;
                }
                if(get_bits_count(&s->gb) >= s->gb.buffer_size*8 && s->in_gb.buffer){
                    skip_bits(&s->in_gb, get_bits_count(&s->gb) - s->gb.buffer_size*8);
                    s->gb= s->in_gb;
                    s->in_gb.buffer=NULL;
                }
                continue;
            }
            bits_pos = get_bits_count(&s->gb);

            if (!s->hdr.lsf) {
                uint8_t *sc;
                int slen, slen1, slen2;

                /* MPEG1 scale factors */
                slen1 = slen_table[0][g->scalefac_compress];
                slen2 = slen_table[1][g->scalefac_compress];
                if (g->block_type == 2) {
                    n = g->switch_point ? 17 : 18;
                    j = 0;
                    if(slen1){
                        for(i=0;i<n;i++)
                            g->scale_factors[j++] = get_bits(&s->gb, slen1);
                    }else{
                        for(i=0;i<n;i++)
                            g->scale_factors[j++] = 0;
                    }
                    if(slen2){
                        for(i=0;i<18;i++)
                            g->scale_factors[j++] = get_bits(&s->gb, slen2);
                        for(i=0;i<3;i++)
                            g->scale_factors[j++] = 0;
                    }else{
                        for(i=0;i<21;i++)
                            g->scale_factors[j++] = 0;
                    }
                } else {
                    sc = s->granules[ch][0].scale_factors;
                    j = 0;
                    for(k=0;k<4;k++) {
                        n = (k == 0 ? 6 : 5);
                        if ((g->scfsi & (0x8 >> k)) == 0) {
                            slen = (k < 2) ? slen1 : slen2;
                            if(slen){
                                for(i=0;i<n;i++)
                                    g->scale_factors[j++] = get_bits(&s->gb, slen);
                            }else{
                                for(i=0;i<n;i++)
                                    g->scale_factors[j++] = 0;
                            }
                        } else {
                            /* simply copy from last granule */
                            for(i=0;i<n;i++) {
                                g->scale_factors[j] = sc[j];
                                j++;
                            }
                        }
                    }
                    g->scale_factors[j++] = 0;
                }
            } else {
                int tindex, tindex2, slen[4], sl, sf;

                /* LSF scale factors */
                if (g->block_type == 2) {
                    tindex = g->switch_point ? 2 : 1;
                } else {
                    tindex = 0;
                }
                sf = g->scalefac_compress;
                if ((s->hdr.mode_ext & MODE_EXT_I_STEREO) && ch == 1) {
                    /* intensity stereo case */
                    sf >>= 1;
                    if (sf < 180) {
                        lsf_sf_expand(slen, sf, 6, 6, 0);
                        tindex2 = 3;
                    } else if (sf < 244) {
                        lsf_sf_expand(slen, sf - 180, 4, 4, 0);
                        tindex2 = 4;
                    } else {
                        lsf_sf_expand(slen, sf - 244, 3, 0, 0);
                        tindex2 = 5;
                    }
                } else {
                    /* normal case */
                    if (sf < 400) {
                        lsf_sf_expand(slen, sf, 5, 4, 4);
                        tindex2 = 0;
                    } else if (sf < 500) {
                        lsf_sf_expand(slen, sf - 400, 5, 4, 0);
                        tindex2 = 1;
                    } else {
                        lsf_sf_expand(slen, sf - 500, 3, 0, 0);
                        tindex2 = 2;
                        g->preflag = 1;
                    }
                }

                j = 0;
                for(k=0;k<4;k++) {
                    n = lsf_nsf_table[tindex2][tindex][k];
                    sl = slen[k];
                    if(sl){
                        for(i=0;i<n;i++)
                            g->scale_factors[j++] = get_bits(&s->gb, sl);
                    }else{
                        for(i=0;i<n;i++)
                            g->scale_factors[j++] = 0;
                    }
                }
                /* XXX: should compute exact size */
                for(;j<40;j++)
                    g->scale_factors[j] = 0;
            }
            exponents_from_scale_factors(s, g, exponents);

            /* read Huffman coded residue */
            huffman_decode(s, g, exponents, bits_pos + g->part2_3_length);
        } /* ch */

        if (s->nb_channels == 2)
            compute_stereo(s, &s->granules[0][gr], &s->granules[1][gr]);

        for(ch=0;ch<s->nb_channels;ch++) {
            g = &s->granules[ch][gr];
            reorder_block(s, g);
            compute_antialias(s, g);
            compute_imdct(s, g, &s->sb_samples[ch][18 * gr][0], s->mdct_buf[ch]);
        }
    } /* gr */
    if(get_bits_count(&s->gb)<0)
        skip_bits(&s->gb, -get_bits_count(&s->gb));
    return nb_granules * 18;
}

static int mp_decode_frame(MPADecodeContext *s, const uint8_t *buf, uint32_t buf_size)
{
    int i, nb_frames, ch;
    float *samples_ptr;

    init_get_bits(&s->gb, buf + HEADER_SIZE, (buf_size - HEADER_SIZE));

    /* skip error protection field (CRC16, polynomial 0x8005) */
    if (s->hdr.error_protection)
        skip_bits(&s->gb, 16);

    //dprintf("frame %d:\n", s->frame_count);
    switch(s->hdr.layer) {
    case 1:
        s->hdr.frame_size = 384;
        nb_frames = mp_decode_layer1(s);
        break;
    case 2:
        s->hdr.frame_size = 1152;
        nb_frames = mp_decode_layer2(s);
        break;
    case 3:
        s->hdr.frame_size = s->hdr.lsf ? 576 : 1152;
    default:
        nb_frames = mp_decode_layer3(s);

        s->last_buf_size=0;
        if(s->in_gb.buffer){
            align_get_bits(&s->gb);
            i= get_bits_left(&s->gb)>>3;
            if(i >= 0 && i <= BACKSTEP_SIZE){
                memmove(s->last_buf, s->gb.buffer + (get_bits_count(&s->gb)>>3), i);
                s->last_buf_size=i;
            }else{
                av_log(AV_LOG_ERROR, "invalid old backstep %d\n", i);
            }
            s->gb= s->in_gb;
            s->in_gb.buffer= NULL;
        }

        align_get_bits(&s->gb);
        i= get_bits_left(&s->gb)>>3;

        if(i<0 || i > BACKSTEP_SIZE || nb_frames<0){
            if(i<0)
                av_log(AV_LOG_ERROR, "invalid new backstep %d\n", i);
            i= FFMIN(BACKSTEP_SIZE, buf_size - HEADER_SIZE);
        }
        memcpy(s->last_buf + s->last_buf_size, s->gb.buffer + buf_size - HEADER_SIZE - i, i);
        s->last_buf_size += i;
        break;
    }

    /* apply the synthesis filter */
    for(ch=0;ch<s->nb_channels;ch++) {
        samples_ptr = s->outbuf[ch];
        for(i=0;i<nb_frames;i++) {
            ff_mpa_synth_filter(s->synth_buf[ch], &(s->synth_buf_offset[ch]),
                                &s->dither_state, samples_ptr, s->sb_samples[ch][i]);
            samples_ptr += 32;
        }
    }

    return nb_frames * 32 * s->nb_channels;
}

int mp3_decode_frame(MPADecodeContext *s, float *data[2], uint32_t *data_size,
                     const uint8_t *buf, uint32_t buf_size)
{
    uint32_t header;
    unsigned int out_size;

    // Discard too short frames
    if (buf_size < HEADER_SIZE) {
        av_log(AV_LOG_ERROR, "Packet is too small\n");
        return -1;
    }

    header = AV_RB32(buf);
    if(ff_mpa_check_header(header) < 0){
        av_log(AV_LOG_ERROR, "Header missing\n");
        return -1;
    }

    if (ff_mpegaudio_decode_header(&s->hdr, header) == 1) {
        /* free format: prepare to compute frame size */
        s->hdr.frame_size = -1;
        return -1;
    }

    s->nb_channels = s->hdr.nb_channels;
#if 0
    if(*data_size < 1152*s->hdr.nb_channels)
        return -1;
#endif
    *data_size = 0;
    if(s->hdr.frame_size<=0 || s->hdr.frame_size > buf_size){
        av_log(AV_LOG_ERROR, "incomplete frame\n");
        return -1;
    }else if(s->hdr.frame_size < buf_size){
        //av_log(AV_LOG_ERROR, "incorrect frame size\n");
        buf_size= s->hdr.frame_size;
    }

    s->adu_mode = 0;
    data[0] = s->outbuf[0];
    data[1] = s->outbuf[1];
    out_size = mp_decode_frame(s, buf, buf_size);
    if(out_size>=0){
        *data_size = out_size;
    }else{
        write(2, "Error while decoding MPEG audio frame.\n", 39); //FIXME return -1 / but also return the number of bytes consumed
    }
    s->hdr.frame_size = 0;
    return buf_size;
}

int mp3_decode_frame_adu(MPADecodeContext *s, float *data[2], uint32_t *data_size,
                         const uint8_t *buf, uint32_t buf_size)
{
    uint32_t header;
    int len = buf_size, out_size;

    // Discard too short frames
    if (buf_size < HEADER_SIZE) {
        av_log(AV_LOG_ERROR, "Packet is too small\n");
        return -1;
    }

    if (len > MPA_MAX_CODED_FRAME_SIZE)
        len = MPA_MAX_CODED_FRAME_SIZE;

    // Get header and restore sync word
    header = AV_RB32(buf) | 0xffe00000;
    if (ff_mpa_check_header(header) < 0) { // Bad header, discard frame
        av_log(AV_LOG_ERROR, "Invalid frame header\n");
        return -1;
    }

    ff_mpegaudio_decode_header(&s->hdr, header);
    s->nb_channels = s->hdr.nb_channels;
    s->hdr.frame_size = len;

    *data_size = 0;
    data[0] = s->outbuf[0];
    data[1] = s->outbuf[1];
    out_size = mp_decode_frame(s, buf, buf_size);
    if(out_size>=0){
        *data_size = out_size;
    }else{
        write(2, "Error while decoding MPEG audio frame.\n", 39); //FIXME return -1 / but also return the number of bytes consumed
    }
    return buf_size;
}

