;*****************************************************************************
;* x86inc.asm
;*****************************************************************************
;* Copyright (C) 2005-2008 x264 project
;*
;* Authors: Loren Merritt <lorenm@u.washington.edu>
;*          Anton Mitrofanov <BugMaster@narod.ru>
;*
;* Permission to use, copy, modify, and/or distribute this software for any
;* purpose with or without fee is hereby granted, provided that the above
;* copyright notice and this permission notice appear in all copies.
;*
;* THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
;* WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
;* MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
;* ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
;* WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
;* ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
;* OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
;*****************************************************************************

; This is a header file for the x264ASM assembly language, which uses
; NASM/YASM syntax combined with a large number of macros to provide easy
; abstraction between different calling conventions (x86_32, win64, linux64).
; It also has various other useful features to simplify writing the kind of
; DSP functions that are most often used in x264.

; Unlike the rest of x264, this file is available under an ISC license, as it
; has significant usefulness outside of x264 and we want it to be available
; to the largest audience possible.  Of course, if you modify it for your own
; purposes to add a new feature, we strongly encourage contributing a patch
; as this feature might be useful for others as well.  Send patches or ideas
; to x264-devel@videolan.org .

%define UNIX64 0
%define WIN64  0
%ifdef ARCH_X86_64
    %ifidn __OUTPUT_FORMAT__,win64
        %define WIN64  1
    %else
        %define UNIX64 1
    %endif
%endif

%ifdef ARCH_X86_64
%define pointer resq
%else
%define pointer resd
%endif

%ifidn __OUTPUT_FORMAT__,elf
%define SIZE_DIRECTIVE_NEEDED
%elifidn __OUTPUT_FORMAT__,elf32
%define SIZE_DIRECTIVE_NEEDED
%elifidn __OUTPUT_FORMAT__,elf64
%define SIZE_DIRECTIVE_NEEDED
%endif

%ifidn __OUTPUT_FORMAT__,win32
%define PREFIX
%endif

%ifdef PREFIX
    %define mangle(x) _ %+ x
%else
    %define mangle(x) x
%endif

; FIXME: All of the 64bit asm functions that take a stride as an argument
; via register, assume that the high dword of that register is filled with 0.
; This is true in practice (since we never do any 64bit arithmetic on strides,
; and x264's strides are all positive), but is not guaranteed by the ABI.

%if WIN64 == 1
    %define PIC
%elifndef ARCH_X86_64
; x86_32 doesn't require PIC.
; Some distros prefer shared objects to be PIC, but nothing breaks if
; the code contains a few textrels, so we'll skip that complexity.
    %undef PIC
%endif
%ifdef PIC
    default rel
%endif

CPU basicnop

; Macros to eliminate most code duplication between x86_32 and x86_64:
; Currently this works only for leaf functions which load all their arguments
; into registers at the start, and make no other use of the stack. Luckily that
; covers most of x264's asm.

; PROLOGUE:
; %1 = number of arguments. loads them from stack if needed.
; %2 = number of registers used. pushes callee-saved regs if needed.
; %3 = number of xmm registers used. pushes callee-saved xmm regs if needed.
; %4 = list of names to define to registers
; PROLOGUE can also be invoked by adding the same options to cglobal

; e.g.
; cglobal foo, 2,3,0, dst, src, tmp
; declares a function (foo), taking two args (dst and src) and one local variable (tmp)

; TODO Some functions can use some args directly from the stack. If they're the
; last args then you can just not declare them, but if they're in the middle
; we need more flexible macro.

; RET:
; Pops anything that was pushed by PROLOGUE

; registers:
; rN and rNq are the native-size register holding function argument N
; rNd, rNw, rNb are dword, word, and byte size
; rNm is the original location of arg N (a register or on the stack), dword
; rNmp is native size

%macro DECLARE_REG 5-6
    %define r%1q %2
    %define r%1d %3
    %define r%1w %4
    %define r%1b %5
    %if %0 == 5
        %ifdef ARCH_X86_64 ; memory
            %define r%1m  %2
        %else
            %define r%1m  %3
        %endif
        %define r%1mp %2
    %elifdef ARCH_X86_64 ; memory
        %define r%1m [rsp + stack_offset + %6]
        %define r%1mp qword r %+ %1m
    %else
        %define r%1m [esp + stack_offset + %6]
        %define r%1mp dword r %+ %1m
    %endif
%endmacro

%macro DECLARE_REG_SIZE 2
    %define r%1q r%1
    %define e%1q r%1
    %define r%1d e%1
    %define e%1d e%1
    %define r%1w %1
    %define e%1w %1
    %define r%1b %2
    %define e%1b %2
%ifndef ARCH_X86_64
    %define r%1  e%1
%endif
%endmacro

DECLARE_REG_SIZE ax, al
DECLARE_REG_SIZE bx, bl
DECLARE_REG_SIZE cx, cl
DECLARE_REG_SIZE dx, dl
DECLARE_REG_SIZE si, sil
DECLARE_REG_SIZE di, dil
DECLARE_REG_SIZE bp, bpl

; t# defines for when per-arch register allocation is more complex than just function arguments

%macro DECLARE_REG_TMP 1-*
    %assign %%i 0
    %rep %0
        CAT_XDEFINE t, %%i, r%1
        %assign %%i %%i+1
        %rotate 1
    %endrep
%endmacro

%macro DECLARE_REG_TMP_SIZE 0-*
    %rep %0
        %define t%1q t%1 %+ q
        %define t%1d t%1 %+ d
        %define t%1w t%1 %+ w
        %define t%1b t%1 %+ b
        %rotate 1
    %endrep
%endmacro

DECLARE_REG_TMP_SIZE 0,1,2,3,4,5,6,7,8,9,10,11,12,13

%ifdef ARCH_X86_64
    %define gprsize 8
%else
    %define gprsize 4
%endif

%macro PUSH 1
    push %1
    %assign stack_offset stack_offset+gprsize
%endmacro

%macro POP 1
    pop %1
    %assign stack_offset stack_offset-gprsize
%endmacro

%macro PUSH_IF_USED 1-*
    %rep %0
        %if %1 < regs_used
            PUSH r%1
        %endif
        %rotate 1
    %endrep
%endmacro

%macro POP_IF_USED 1-*
    %rep %0
        %if %1 < regs_used
            pop r%1
        %endif
        %rotate 1
    %endrep
%endmacro

%macro LOAD_IF_USED 1-*
    %rep %0
        %if %1 < num_args
            mov r%1, r %+ %1 %+ mp
        %endif
        %rotate 1
    %endrep
%endmacro

%macro SUB 2
    sub %1, %2
    %ifidn %1, rsp
        %assign stack_offset stack_offset+(%2)
    %endif
%endmacro

%macro ADD 2
    add %1, %2
    %ifidn %1, rsp
        %assign stack_offset stack_offset-(%2)
    %endif
%endmacro

%macro movifnidn 2
    %ifnidn %1, %2
        mov %1, %2
    %endif
%endmacro

%macro movsxdifnidn 2
    %ifnidn %1, %2
        movsxd %1, %2
    %endif
%endmacro

%macro ASSERT 1
    %if (%1) == 0
        %error assert failed
    %endif
%endmacro

%macro DEFINE_ARGS 0-*
    %ifdef n_arg_names
        %assign %%i 0
        %rep n_arg_names
            CAT_UNDEF arg_name %+ %%i, q
            CAT_UNDEF arg_name %+ %%i, d
            CAT_UNDEF arg_name %+ %%i, w
            CAT_UNDEF arg_name %+ %%i, b
            CAT_UNDEF arg_name %+ %%i, m
            CAT_UNDEF arg_name, %%i
            %assign %%i %%i+1
        %endrep
    %endif

    %assign %%i 0
    %rep %0
        %xdefine %1q r %+ %%i %+ q
        %xdefine %1d r %+ %%i %+ d
        %xdefine %1w r %+ %%i %+ w
        %xdefine %1b r %+ %%i %+ b
        %xdefine %1m r %+ %%i %+ m
        CAT_XDEFINE arg_name, %%i, %1
        %assign %%i %%i+1
        %rotate 1
    %endrep
    %assign n_arg_names %%i
%endmacro

%if WIN64 ; Windows x64 ;=================================================

DECLARE_REG 0,  rcx, ecx,  cx,   cl
DECLARE_REG 1,  rdx, edx,  dx,   dl
DECLARE_REG 2,  r8,  r8d,  r8w,  r8b
DECLARE_REG 3,  r9,  r9d,  r9w,  r9b
DECLARE_REG 4,  r10, r10d, r10w, r10b, 40
DECLARE_REG 5,  r11, r11d, r11w, r11b, 48
DECLARE_REG 6,  rax, eax,  ax,   al,   56
DECLARE_REG 7,  rbx, ebx,  bx,   bl,   64
DECLARE_REG 8,  rdi, edi,  di,   dil,  72
DECLARE_REG 9,  rsi, esi,  si,   sil,  80
DECLARE_REG 10, r12, r12d, r12w, r12b, 88
DECLARE_REG 11, r13, r13d, r13w, r13b, 96
DECLARE_REG 12, r14, r14d, r14w, r14b, 104
DECLARE_REG 13, r15, r15d, r15w, r15b, 112
%define r0 rcx
%define r1 rdx
%define r6 rax
%define r7 rbx 

%macro PROLOGUE 2-4+ 0 ; #args, #regs, #xmm_regs, arg_names...
    %assign num_args %1
    %assign regs_used %2
    ASSERT regs_used >= num_args
    ASSERT regs_used <= 14
    PUSH_IF_USED 7, 8, 9, 10, 11, 12, 13
    WIN64_SPILL_XMM %3
    LOAD_IF_USED 4, 5, 6, 7, 8, 9, 10, 11, 12, 13
    DEFINE_ARGS %4
%endmacro

%macro WIN64_SPILL_XMM 1
    %assign xmm_regs_used %1
    ASSERT xmm_regs_used <= 16
    ; Use the shadow space to store XMM6 and XMM7, the rest needs stack space allocated.
    %if xmm_regs_used > 6
        movaps [rsp + stack_offset +  8], xmm6
    %endif
    %if xmm_regs_used > 7
        movaps [rsp + stack_offset + 24], xmm7
    %endif
    %if xmm_regs_used > 8
        sub rsp, (xmm_regs_used-8)*16+48
        %assign stack_offset stack_offset+(xmm_regs_used-8)*16+48
        %assign %%i xmm_regs_used
        %rep (xmm_regs_used-8)
            %assign %%i %%i-1
            movaps [rsp + (%%i-8)*16+48], xmm %+ %%i
        %endrep
    %endif
%endmacro

%macro WIN64_RESTORE_XMM_INTERNAL 1
    %if xmm_regs_used > 8 
        %assign %%i xmm_regs_used
        %rep (xmm_regs_used-8)
            %assign %%i %%i-1
            movaps xmm %+ %%i, [%1 + (%%i-8)*16+48]
        %endrep
        add %1, (xmm_regs_used-8)*16+48
    %endif
    %if xmm_regs_used > 7
        movaps xmm7, [%1 + stack_offset + 24]
    %endif
    %if xmm_regs_used > 6
        movaps xmm6, [%1 + stack_offset +  8]
    %endif
%endmacro

%macro WIN64_RESTORE_XMM 1
    WIN64_RESTORE_XMM_INTERNAL %1
    %assign stack_offset stack_offset-(xmm_regs_used-8)*16+16
    %assign xmm_regs_used 0
%endmacro

%macro RET 0
    WIN64_RESTORE_XMM_INTERNAL rsp
    POP_IF_USED 13, 12, 11, 10, 9, 8, 7
%if avx_enabled
    vzeroupper
%endif
    pop rbp
    ret
%endmacro

%elifdef ARCH_X86_64 ; *nix x64 ;=============================================

DECLARE_REG 0, rdi, edi, di,  dil
DECLARE_REG 1, rsi, esi, si,  sil
DECLARE_REG 2, rdx, edx, dx,  dl
DECLARE_REG 3, rcx, ecx, cx,  cl
DECLARE_REG 4, r8,  r8d, r8w, r8b
DECLARE_REG 5, r9,  r9d, r9w, r9b
DECLARE_REG 6,  rax, eax,  ax,   al,  8
DECLARE_REG 7,  rbx, ebx,  bx,   bl, 16
DECLARE_REG 8,  r10, r10d, r10w, r10b, 24
DECLARE_REG 9,  r11, r11d, r11w, r11b, 32
DECLARE_REG 10, r12, r12d, r12w, r12b, 40
DECLARE_REG 11, r13, r13d, r13w, r13b, 48
DECLARE_REG 12, r14, r14d, r14w, r14b, 56
DECLARE_REG 13, r15, r15d, r15w, r15b, 64
%define r0 rdi
%define r1 rsi
%define r2 rdx
%define r3 rcx
%define r6 rax
%define r7 rbx

%macro PROLOGUE 2-4+ ; #args, #regs, #xmm_regs, arg_names...
    %assign num_args %1
    %assign regs_used %2
    ASSERT regs_used >= num_args
    ASSERT regs_used <= 14
    PUSH_IF_USED 9, 10, 11, 12, 13
    LOAD_IF_USED 6, 7, 8, 9, 10, 11, 12, 13
    DEFINE_ARGS %4
%endmacro

%macro RET 0
    POP_IF_USED 13, 12, 11, 10, 9
%if avx_enabled
    vzeroupper
%endif
    pop rbp
    ret
%endmacro

%else ; X86_32 ;==============================================================

DECLARE_REG 0, eax, eax, ax, al,   4
DECLARE_REG 1, ecx, ecx, cx, cl,   8
DECLARE_REG 2, edx, edx, dx, dl,   12
DECLARE_REG 3, ebx, ebx, bx, bl,   16
DECLARE_REG 4, esi, esi, si, null, 20
DECLARE_REG 5, edi, edi, di, null, 24
%define rbp ebp
%define rsp esp

%macro DECLARE_ARG 1-*
    %rep %0
        %define r%1m [esp + stack_offset + 4*%1 + 4]
        %define r%1mp dword r%1m
        %rotate 1
    %endrep
%endmacro

DECLARE_ARG 6, 7, 8, 9, 10, 11, 12, 13

%macro PROLOGUE 2-4+ ; #args, #regs, #xmm_regs, arg_names...
    %assign num_args %1
    %assign regs_used %2
    ASSERT regs_used <= 6
    PUSH_IF_USED 3, 4, 5
    LOAD_IF_USED 0, 1, 2, 3, 4, 5
    DEFINE_ARGS %4
%endmacro

%macro RET 0
    POP_IF_USED 5, 4, 3
%if avx_enabled
    vzeroupper
%endif
    pop rbp
    ret
%endmacro

%endif ;======================================================================

%if WIN64 == 0
%macro WIN64_SPILL_XMM 1
%endmacro
%macro WIN64_RESTORE_XMM 1
%endmacro
%endif

;=============================================================================
; arch-independent part
;=============================================================================

%assign function_align 16

; Symbol prefix for C linkage
%macro cglobal 1-2+ "" ; name, [PROLOGUE args]
    cglobal_internal %1 %+ SUFFIX, %2
%endmacro
%macro cglobal_internal 1-2+
    %xdefine %1 mangle(%1)
    %xdefine %1.skip_prologue %1 %+ .skip_prologue
    global %1
;%ifdef SIZE_DIRECTIVE_NEEDED
;    [type %1 function]
;%endif
    align function_align
    %1:
    INIT_XMM ; not really needed, but makes disassembly somewhat nicer
    %assign stack_offset 0
    push rbp
    mov rbp, rsp
    %if %0 > 1
        PROLOGUE %2
    %endif
%endmacro
%macro cendfunc 1
    cendfunc_internal %1 %+ SUFFIX
%endmacro
%macro cendfunc_internal 1
%ifdef SIZE_DIRECTIVE_NEEDED
%1 %+ size EQU $-%1
[size %1 %1 %+ size]
%endif
%endmacro

%macro cextern 1
    extern %1
%endmacro

%macro const 2+
    global %1
    %1: %2
%endmacro

; This is needed for ELF, otherwise the GNU linker assumes the stack is
; executable by default.
%ifidn __OUTPUT_FORMAT__,elf
SECTION .note.GNU-stack noalloc noexec nowrite progbits
%endif

; merge mmx and sse*

%macro CAT_XDEFINE 3
    %xdefine %1%2 %3
%endmacro

%macro CAT_UNDEF 2
    %undef %1%2
%endmacro

%macro INIT_XMM 0
    %assign avx_enabled 0
    %define num_mmregs 8
    %ifdef ARCH_X86_64
    %define num_mmregs 16
    %endif
    %assign %%i 0
    %rep num_mmregs
    CAT_XDEFINE m, %%i, xmm %+ %%i
    CAT_XDEFINE nxmm, %%i, %%i
    %assign %%i %%i+1
    %endrep
%endmacro

%macro INIT_YMM 0-1+
    %assign avx_enabled 1
    %define RESET_MM_PERMUTATION INIT_YMM %1
    %define num_mmregs 8
    %ifdef ARCH_X86_64
    %define num_mmregs 16
    %endif
    %assign %%i 0
    %rep num_mmregs
    CAT_XDEFINE m, %%i, ymm %+ %%i
    CAT_XDEFINE nymm, %%i, %%i
    %assign %%i %%i+1
    %endrep
%endmacro

INIT_XMM

; I often want to use macros that permute their arguments. e.g. there's no
; efficient way to implement butterfly or transpose or dct without swapping some
; arguments.
;
; I would like to not have to manually keep track of the permutations:
; If I insert a permutation in the middle of a function, it should automatically
; change everything that follows. For more complex macros I may also have multiple
; implementations, e.g. the SSE2 and SSSE3 versions may have different permutations.
;
; Hence these macros. Insert a PERMUTE or some SWAPs at the end of a macro that
; permutes its arguments. It's equivalent to exchanging the contents of the
; registers, except that this way you exchange the register names instead, so it
; doesn't cost any cycles.

%macro PERMUTE 2-* ; takes a list of pairs to swap
%rep %0/2
    %xdefine tmp%2 m%2
    %xdefine ntmp%2 nm%2
    %rotate 2
%endrep
%rep %0/2
    %xdefine m%1 tmp%2
    %xdefine nm%1 ntmp%2
    %undef tmp%2
    %undef ntmp%2
    %rotate 2
%endrep
%endmacro

%macro SWAP 2-* ; swaps a single chain (sometimes more concise than pairs)
%rep %0-1
%ifdef m%1
    %xdefine tmp m%1
    %xdefine m%1 m%2
    %xdefine m%2 tmp
    CAT_XDEFINE n, m%1, %1
    CAT_XDEFINE n, m%2, %2
%else
    ; If we were called as "SWAP m0,m1" rather than "SWAP 0,1" infer the original numbers here.
    ; Be careful using this mode in nested macros though, as in some cases there may be
    ; other copies of m# that have already been dereferenced and don't get updated correctly.
    %xdefine %%n1 n %+ %1
    %xdefine %%n2 n %+ %2
    %xdefine tmp m %+ %%n1
    CAT_XDEFINE m, %%n1, m %+ %%n2
    CAT_XDEFINE m, %%n2, tmp
    CAT_XDEFINE n, m %+ %%n1, %%n1
    CAT_XDEFINE n, m %+ %%n2, %%n2
%endif
    %undef tmp
    %rotate 1
%endrep
%endmacro

; If SAVE_MM_PERMUTATION is placed at the end of a function and given the
; function name, then any later calls to that function will automatically
; load the permutation, so values can be returned in mmregs.
%macro SAVE_MM_PERMUTATION 1 ; name to save as
    %assign %%i 0
    %rep num_mmregs
    CAT_XDEFINE %1_m, %%i, m %+ %%i
    %assign %%i %%i+1
    %endrep
%endmacro

%macro LOAD_MM_PERMUTATION 1 ; name to load from
    %assign %%i 0
    %rep num_mmregs
    CAT_XDEFINE m, %%i, %1_m %+ %%i
    CAT_XDEFINE n, m %+ %%i, %%i
    %assign %%i %%i+1
    %endrep
%endmacro

%macro call 1
    call %1
    %ifdef %1_m0
        LOAD_MM_PERMUTATION %1
    %endif
%endmacro

%macro SBUTTERFLYPS 3
    unpckhps m%3, m%1, m%2
    unpcklps m%1, m%1, m%2
%endmacro

%macro DEINTCPLXPS 3
    movaps   %3, %1
    shufps   %3, %2, 0xdd ; m3: [r0 r1 r2 r3]
    shufps   %1, %2, 0x88 ; m1: [i0 i1 i2 i3]
%endmacro

%macro PALIGNR_MMX 4 ; dst, src, imm, tmp
%ifnidn %4, %2
    movaps  %4, %2
%endif
    pslldq  %1, 16-%3
    psrldq  %4, %3
    por     %1, %4
%endmacro

%macro PALIGNR_SSSE3 4
    palignr %1, %2, %3
%endmacro

; base-4 constants for shuffles
%assign i 0
%rep 256
    %assign j ((i>>6)&3)*1000 + ((i>>4)&3)*100 + ((i>>2)&3)*10 + (i&3)
    %if j < 10
        CAT_XDEFINE q000, j, i
    %elif j < 100
        CAT_XDEFINE q00, j, i
    %elif j < 1000
        CAT_XDEFINE q0, j, i
    %else
        CAT_XDEFINE q, j, i
    %endif
%assign i i+1
%endrep
%undef i
%undef j

;=============================================================================
; AVX abstraction layer
;=============================================================================

%assign i 0
%rep 16
    %if i < 8
        CAT_XDEFINE sizeofmm, i, 8
    %endif
    CAT_XDEFINE sizeofxmm, i, 16
    CAT_XDEFINE sizeofymm, i, 32
%assign i i+1
%endrep
%undef i

%macro CHECK_AVX_INSTR_EMU 3-*
    %xdefine %%opcode %1
    %xdefine %%dst %2
    %rep %0-2
        %ifidn %%dst, %3
            %error non-avx emulation of ``%%opcode'' is not supported
        %endif
        %rotate 1
    %endrep
%endmacro

;%1 == instruction
;%2 == unused
;%3 == 1 if 4-operand (xmm, xmm, xmm, imm), 0 if 2- or 3-operand (xmm, xmm, xmm)
;%4 == number of operands given
;%5+: operands
%macro RUN_AVX_INSTR 6-7+
    %ifid %6
        %define %%sizeofreg sizeof%6
    %elifid %5
        %define %%sizeofreg sizeof%5
    %else
%if avx_enabled
        %define %%sizeofreg 32
%else
        %define %%sizeofreg 32
%endif
    %endif
    %if %%sizeofreg==32
        %if %4>=3
            v%1 %5, %6, %7
        %else
            v%1 %5, %6
        %endif
    %else
        %if %4>=3+%3
            %ifnidn %5, %6
                %if avx_enabled && %%sizeofreg==16
                    v%1 %5, %6, %7
                %elifdef vex_encode
                    v%1 %5, %6, %7
                %else
                    movaps %5, %6
                    %1 %5, %7
                %endif
            %else
                %ifdef vex_encode
                    v%1 %5, %5, %7
                %else
                    %1 %5, %7
                %endif
            %endif
        %elif %4>=3
            %1 %5, %6, %7
        %else
            %1 %5, %6
        %endif
    %endif
%endmacro

; 3arg AVX ops with a memory arg can only have it in src2,
; whereas SSE emulation of 3arg prefers to have it in src1 (i.e. the mov).
; So, if the op is symmetric and the wrong one is memory, swap them.
%macro RUN_AVX_INSTR1 8
    %assign %%swap 0
    %if avx_enabled
        %ifnid %6
            %assign %%swap 1
        %endif
    %elifnidn %5, %6
        %ifnid %7
            %assign %%swap 1
        %endif
    %endif
    %if %%swap && %3 == 0 && %8 == 1
        RUN_AVX_INSTR %1, %2, %3, %4, %5, %7, %6
    %else
        RUN_AVX_INSTR %1, %2, %3, %4, %5, %6, %7
    %endif
%endmacro

;%1 == instruction
;%2 == 1 if float, 0 if int
;%3 == 1 if 4-operand (xmm, xmm, xmm, imm), 0 if 2- or 3-operand (xmm, xmm, xmm)
;%4 == 1 if symmetric (i.e. doesn't matter which src arg is which), 0 if not
%macro AVX_INSTR 3
    %macro %1 2-9 fnord, fnord, fnord, %1, %2, %3, %4
        %ifidn %3, fnord
            RUN_AVX_INSTR %6, 0, %7, 2, %1, %2
        %elifidn %4, fnord
            RUN_AVX_INSTR1 %6, 0, %7, 3, %1, %2, %3, %8
        %elifidn %5, fnord
            RUN_AVX_INSTR %6, 0, %7, 4, %1, %2, %3, %4
        %else
            RUN_AVX_INSTR %6, 0, %7, 5, %1, %2, %3, %4, %5
        %endif
    %endmacro
%endmacro

AVX_INSTR addps, 0, 1
AVX_INSTR addss, 0, 1
AVX_INSTR addsubps, 0, 0
AVX_INSTR andps, 0, 1
AVX_INSTR andnps, 0, 0
AVX_INSTR blendps, 0, 0
AVX_INSTR blendvps, 0, 0
AVX_INSTR cmpps, 0, 0
AVX_INSTR cmpss, 0, 0
AVX_INSTR comiss, 0, 0
AVX_INSTR cvtdq2ps, 0, 0
AVX_INSTR cvtps2dq, 0, 0
AVX_INSTR cvttps2dq, 0, 0
AVX_INSTR cvtsi2ss, 0, 0
AVX_INSTR cvtss2si, 0, 0
AVX_INSTR cvttss2si, 0, 0
AVX_INSTR extractps, 0, 0
AVX_INSTR haddps, 0, 0
AVX_INSTR hsubps, 0, 0
AVX_INSTR maxps, 0, 1
AVX_INSTR maxss, 0, 1
AVX_INSTR minps, 0, 1
AVX_INSTR minss, 0, 1
AVX_INSTR movhlps, 0, 0
AVX_INSTR movlhps, 0, 0
AVX_INSTR movmskps, 0, 0
AVX_INSTR movss, 0, 0
AVX_INSTR mulps, 0, 1
AVX_INSTR mulss, 0, 1
AVX_INSTR orps, 0, 1
AVX_INSTR rcpps, 0, 0
AVX_INSTR rcpss, 0, 0
AVX_INSTR rsqrtps, 0, 0
AVX_INSTR rsqrtss, 0, 0
AVX_INSTR sqrtps, 0, 0
AVX_INSTR sqrtss, 0, 0
AVX_INSTR subps, 0, 0
AVX_INSTR subss, 0, 0
AVX_INSTR ucomiss, 0, 0
AVX_INSTR unpckhps, 0, 0
AVX_INSTR unpcklps, 0, 0
AVX_INSTR xorps, 0, 1

AVX_INSTR packsswb, 0, 0
AVX_INSTR packssdw, 0, 0                                                                                                                                          
AVX_INSTR packuswb, 0, 0
AVX_INSTR packusdw, 0, 0
AVX_INSTR paddb, 0, 1
AVX_INSTR paddw, 0, 1
AVX_INSTR paddd, 0, 1
AVX_INSTR paddq, 0, 1
AVX_INSTR paddsb, 0, 1
AVX_INSTR paddsw, 0, 1
AVX_INSTR paddusb, 0, 1
AVX_INSTR paddusw, 0, 1
AVX_INSTR pand, 0, 1
AVX_INSTR pandn, 0, 0
AVX_INSTR pavgb, 0, 1
AVX_INSTR pavgw, 0, 1
AVX_INSTR pblendvb, 0, 0
AVX_INSTR pcmpeqb, 0, 1
AVX_INSTR pcmpeqw, 0, 1
AVX_INSTR pcmpeqd, 0, 1
AVX_INSTR pcmpeqq, 0, 1
AVX_INSTR pcmpgtb, 0, 0
AVX_INSTR pcmpgtw, 0, 0
AVX_INSTR pcmpgtd, 0, 0
; SSE4.2!
AVX_INSTR pcmpgtq, 0, 0
AVX_INSTR pmaddwd, 0, 1
AVX_INSTR pmaxsb, 0, 1
AVX_INSTR pmaxsw, 0, 1
AVX_INSTR pmaxsd, 0, 1
AVX_INSTR pmaxub, 0, 1
AVX_INSTR pmaxuw, 0, 1
AVX_INSTR pmaxud, 0, 1
AVX_INSTR pminsb, 0, 1
AVX_INSTR pminsw, 0, 1
AVX_INSTR pminsd, 0, 1
AVX_INSTR pminub, 0, 1
AVX_INSTR pminuw, 0, 1
AVX_INSTR pminud, 0, 1
AVX_INSTR pmuldq, 0, 1
AVX_INSTR pmulhrsw, 0, 1
AVX_INSTR pmulhuw, 0, 1
AVX_INSTR pmulhw, 0, 1
AVX_INSTR pmullw, 0, 1
AVX_INSTR pmulld, 0, 1
AVX_INSTR pmuludq, 0, 1
AVX_INSTR por, 0, 1
AVX_INSTR psadbw, 0, 1
AVX_INSTR psllw, 0, 0
AVX_INSTR pslld, 0, 0
AVX_INSTR psllq, 0, 0
AVX_INSTR pslldq, 0, 0
AVX_INSTR psraw, 0, 0
AVX_INSTR psrad, 0, 0
AVX_INSTR psrlw, 0, 0
AVX_INSTR psrld, 0, 0
AVX_INSTR psrlq, 0, 0
AVX_INSTR psrldq, 0, 0
AVX_INSTR psubb, 0, 0
AVX_INSTR psubw, 0, 0
AVX_INSTR psubd, 0, 0
AVX_INSTR psubq, 0, 0
AVX_INSTR psubsb, 0, 0
AVX_INSTR psubsw, 0, 0
AVX_INSTR psubusb, 0, 0
AVX_INSTR psubusw, 0, 0
AVX_INSTR punpckhbw, 0, 0
AVX_INSTR punpckhwd, 0, 0
AVX_INSTR punpckhdq, 0, 0
AVX_INSTR punpckhqdq, 0, 0
AVX_INSTR punpcklbw, 0, 0
AVX_INSTR punpcklwd, 0, 0
AVX_INSTR punpckldq, 0, 0
AVX_INSTR punpcklqdq, 0, 0
AVX_INSTR pxor, 0, 1

AVX_INSTR dpps, 1, 0
AVX_INSTR insertps, 1, 0
AVX_INSTR shufps, 1, 0

